import voxelfarm

class generator_lambda_framework:

    def __init__(self):  
       pass

    def input(self, id, label, default = 0.0):
        return 0.0

    def set_vector(self, index, vx, vy, vz):
        pass

    def set_material(self, index, id):
        pass

    def set_field(self, index, field):
        pass

    def set_entity_bounds(self, axis, v0, v1):
        pass

    def get_voxel_origin(self, index, dimension):
        return (0,0,0)

    def get_voxel_origin_axis(self, index, dimension, axis):
        return (0,0,0)

    def get_field_origin(self, index, dimension):
        return (0,0,0)

    def test(self):
        pass

class generator_lambda_host:

    version = (3,0,1,7)

    voxels = range(0, 44 * 44 * 44)
    field = range(0, 45 * 45 * 45)

    def __init__(self, framework = None):
        if framework:
            self.lambda_framework = framework
        else:
            if voxelfarm.voxelfarm_framework:
                self.lambda_framework = voxelfarm.voxelfarm_framework
            else:
                self.lambda_framework = generator_lambda_framework()

    def set_vector(self, index, vx, vy, vz):
        self.lambda_framework.set_vector(index, vx, vy, vz)

    def set_material(self, index, id):
        self.lambda_framework.set_material(index, id)

    def set_field(self, index, field):
        self.lambda_framework.set_field(index, field)

    def get_voxel_origin(self, index):
        return self.lambda_framework.get_voxel_origin(index, 44)

    def get_field_origin(self, index):
        return self.lambda_framework.get_voxel_origin(index, 45)

    def get_field_origin_x(self, index):
        return self.lambda_framework.get_voxel_origin_axis(index, 45, 0)

    def get_field_origin_y(self, index):
        return self.lambda_framework.get_voxel_origin_axis(index, 45, 1)

    def get_field_origin_z(self, index):
        return self.lambda_framework.get_voxel_origin_axis(index, 45, 2)

    def input(self, id, label, default = 0.0):
        return self.lambda_framework.input(id, label, default)

    def set_entity_bounds_x(self, v0, v1):
        if self.lambda_framework.set_entity_bounds(0, v0, v1) == 0:
            self.cellIsEmpty()

    def set_entity_bounds_y(self, v0, v1):
        if self.lambda_framework.set_entity_bounds(1, v0, v1) == 0:
            self.cellIsEmpty()

    def set_entity_bounds_z(self, v0, v1):
        if self.lambda_framework.set_entity_bounds(2, v0, v1) == 0:
            self.cellIsEmpty()

    def cellIsEmpty(self):
        self.voxels = range(0, 0)
        self.field = range(0, 0)

    def test(self):
        pass



from datetime import datetime

voxelfarm_framework = None

class property_buffer_type:
    Buffer = 'buffer'
    Dataframe = 'dataframe'
    Voxel = 'voxels'

class special_properties:
    Merge = 'voxelfarm_merge'
    BufferType = 'voxelfarm_buffer_type'
    Compress = 'voxelfarm_compress'
    FixSelfIntersections = 'voxelfarm_fix_self_intersections'

class entity_type:
    View = 'VIEW'
    Project = 'PROJECT'
    VoxelTerrain = 'VOXSURF'
    RealtimeVoxelTerrain = 'VOXOP'
    BlockModel = 'VOXBM'
    IndexedPointCloud = 'IDXPC'
    VoxelGenerator = 'VOXGEN'
    RawPointCloud = 'RAWPC'
    RawHeightmap = 'RAWHM'
    RawBlockModel = 'RAWBM'
    RawMesh = 'RAWMESH'
    IndexedMesh = 'IDXMESH'
    VoxelMesh = 'VOXMESH'
    OrthoImagery = 'ORTHO'
    IndexedOrthoImagery = 'IDXORTHO'
    Program = 'PROGRAM'
    Folder = 'FOLDER'
    RawDensity = 'RAWDENSITY'
    IndexedDensity = 'IDXDENSITY'
    MaterialTracking = 'VOXMT'
    MaterialTrackingOperation = 'VOXMTOP'
    RawDrillHoles = 'RAWDH'
    DrillHoles = 'IDXDH'
    OrthoVox='ORTHOVOX'
    Process='PROCESS'
    Export='EXPORT'
    Report = 'REPORT'
    VoxelPC = 'VOXPC'
    RawGeoChem = 'RAWGEOCHEM'
    GeoChem = 'GEOCHEM'
    TerrainComposite = 'SURFCOMPOSITE'

class lambda_type:
    Generator = 'VOXEL'
    Report = 'REPORT'
    View = 'VIEW'
    Workflow = 'WORKFLOW'
    Process = 'PROCESS'
    
class process_lambda_type:
    Workflow = 'WORKFLOW'
    Process = 'PROCESS'

class operation_type:
    sum = 0
    min = 1
    max = 2
    avg = 3

class type:
    view = 1 << 1
    project = 1 << 2
    voxel_terrain = 1 << 3
    voxel_operations = 1 << 4
    block_model = 1 << 5
    point_cloud = 1 << 6
    voxel_generator = 1 << 7
    point_cloud_raw = 1 << 8
    heightmap_raw = 1 << 9
    block_model_raw = 1 << 10
    mesh_raw = 1 << 11
    mesh = 1 << 12
    ortho = 1 << 13
    program = 1 << 14
    folder = 1 << 15
    raw_density = 1 << 16
    indexed_density = 1 << 17
    material_tracking = 1 << 18
    voxel_mesh = 1 << 19
    drill_holes = 1 << 20
    drill_holes_raw = 1 << 21
    indexed_ortho = 1 << 22
    voxel_points = 1 << 23
    voxel_plane = 1 << 24
    wms_ortho = 1 << 25
    voxel_ortho = 1 << 26
    export = 1 << 27
    report = 1 << 28
    raw_geo_chem = 1 << 29
    geo_chem = 1 << 30
    wmts_ortho = 1 << 31
    terrain_composite = 1 << 32
    value = 1
    set = 2

class attribute:
    none = 0
    volume = 1
    material = 1 << 1
    color = 1 << 2
    vector = 1 << 3
    normal = 1 << 4
    tangent = 1 << 5
    uv = 1 << 6
    color_difference = 1 << 7
    voxel_index = 1 << 8
    faces = 1 << 9

class attribute_merge_mode:
    none = 0
    add = 1
    multiply = 2
    min = 3
    max = 4
    average = 5
    unit = 6
    grade = 7

class layer_merge_mode:
    none = 0
    min = 1
    max = 2
    override = 3

class texture:
    diffuse = 0
    normal = 1
    rgb0 = 2
    rgb1 = 3
    rgb2 = 4
    rgba0 = 1
    rgba1 = 1
    rgba2 = 1

class ColorLegend:
    gradient_step = 1.0
    gradient_min = 0.0
    gradient_max = 100.0
    gradient_attribute = ""
    static_color = (0, 0, 0)

class DrillHoleOptions:
    show_collars = True
    show_cylinders = True

class Material:
    def __init__(self):  
        self.color = (1.0, 1.0, 1.0, 1.0)
        self.gradient = None
        self.diffuse = False
        self.normal = False
        self.gradient = ""
        self.simplification_error = 0.00005
    
    def serialize(self):
        result = "0,"
        result += str(self.color[0]) + ","
        result += str(self.color[1]) + ","
        result += str(self.color[2]) + ","
        result += str(self.color[3]) + ","
        result += ("1" if self.diffuse else "0") + ","
        result += ("1" if self.normal else "0") + ","
        result += self.gradient + ","
        result += str(self.simplification_error) 
        return result

version = (3,0,1,7)

# Reports and Views

voxelCount = 40*40*40
fieldCount = 45*45*45

voxels = range(0, voxelCount)
field = range(0, fieldCount)

def serialize_attributes(attributes):
    attribute_list = ''
    if attributes:
        if isinstance(attributes, str):
            attribute_list = attributes
        else:    
            attribute_list = ','.join(attributes)

    return attribute_list

def get_voxel_size():
    if voxelfarm_framework:
        return voxelfarm_framework.get_voxel_size()
    return 1.0

def input(id, label, default = 0.0):
    if voxelfarm_framework:
        return voxelfarm_framework.input(id, label, default)
    return default

def input_entity(id, label, type):
    if voxelfarm_framework:
        return voxelfarm_framework.input_entity(id, label, type)
    return -1

def input_attributes(id, label, type, entity):
    if voxelfarm_framework:
        return voxelfarm_framework.input_attributes(id, label, type, entity)
    return ''

def input_attribute(id, label, type, entity):
    if voxelfarm_framework:
        return voxelfarm_framework.input_attributes(id, label, type, entity)
    return ''

def input_query(id, label, entity):
    if voxelfarm_framework:
        return voxelfarm_framework.input_query(id, label, entity)
    return ''

def input_region(id, label):
    if voxelfarm_framework:
        return voxelfarm_framework.input_region(id, label)
    return ''

def input_color_legend(attributes, default_value):
    result = ColorLegend()
    gradient_step = 1.0
    gradient_min = 0.0
    gradient_max = 100.0
    gradient_attribute = "fe"
    if default_value != None:
        gradient_step = default_value.gradient_step
        gradient_min = default_value.gradient_min
        gradient_max = default_value.gradient_max
        gradient_attribute = default_value.gradient_attribute

    if voxelfarm_framework:
        legend_string = voxelfarm_framework.input_color_legend(attributes, gradient_step, gradient_min, gradient_max, gradient_attribute)
        items = legend_string.split(',')
        result.gradient_attribute = items[0]
        result.gradient_min = float(items[1])
        result.gradient_max = float(items[2])
        result.gradient_step = float(items[3])
        return result

    result.gradient_attribute = gradient_attribute
    result.gradient_min = gradient_min
    result.gradient_max = gradient_max
    result.gradient_step = gradient_step
    return result
    
def input_drillhole_options(entity, default):
    result = DrillHoleOptions()
    result.show_collars = 1
    result.show_cylinders = 1
    if default != None:
        result.show_collars = 1 if default.show_collars else 0
        result.show_cylinders = 1 if default.show_cylinders else 0

    if voxelfarm_framework:
        voxelfarm_framework.input_drillhole_options(entity, result.show_collars, result.show_cylinders)
    return result

def input_date(id, label, default = datetime.now):
    if voxelfarm_framework:
        timestamp = voxelfarm_framework.input_date(id, label, 0)
        return datetime.fromtimestamp(timestamp/1000) # we lose the milliseconds
    return default    

def input_string(id, label, default = ""):
    if voxelfarm_framework:
        return voxelfarm_framework.input_string(id, label, default)
    return default    

def input_bool(id, label, default = None):
    if voxelfarm_framework:
        return voxelfarm_framework.input_bool(id, label, default)
    return default

def set_query(entity, query):
    if voxelfarm_framework:
        voxelfarm_framework.set_query(entity, query)

def set_attributes(entity, attributes):
    if voxelfarm_framework:
        voxelfarm_framework.set_attributes(entity, attributes)

def set_attribute_gradient(entity, attribute, min_value, max_value):
    if voxelfarm_framework:
        voxelfarm_framework.set_attribute_gradient(entity, attribute, min_value, max_value)

def start_composite():
    if voxelfarm_framework:
        return voxelfarm_framework.start_composite()
    return -1

def add_layer(layers, entity, merge_mode=layer_merge_mode.override):
    if voxelfarm_framework:
        voxelfarm_framework.add_layer(layers, entity, merge_mode)

def get_entity_type(entity):
    if voxelfarm_framework:
        return voxelfarm_framework.get_entity_type(entity)
    return 0

def get_entity_name(entity):
    if voxelfarm_framework:
        return voxelfarm_framework.get_entity_name(entity)
    return ''

def get_entity_date(entity):
    if voxelfarm_framework:
        timestamp = voxelfarm_framework.get_entity_date(entity)
        return datetime.fromtimestamp(timestamp/1000) #we lose the milliseconds
    return datetime.now

def get_entity_attributes(entity):
    if voxelfarm_framework:
        attributes = voxelfarm_framework.get_entity_attributes(entity)
        return attributes
    return ''

def folder_contains(folder, entity):
    if voxelfarm_framework:
        return voxelfarm_framework.folder_contains(folder, entity)
    return False

def get_attributes(attributes):
    if voxelfarm_framework:
        count = voxelfarm_framework.get_attribute_count(attributes)
        return range(0, count)

    attrs = attribute.split()   
    return range(0, len(attrs))      

def get_attribute_name(attributes, index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_attribute_name(attributes, index)

    attrs = attribute.split()   
    if index < len(attrs):
        return attrs[index]
    return ''

def get_attribute_index(attributes, name):
    if voxelfarm_framework:
        return voxelfarm_framework.get_attribute_index(attributes, name)

    attrs = attribute.split()   
    if name in attrs:
        return attrs.index(name)
    return -1    

def start_sum(item, property):
    return start_operation(item, property, operation_type.sum)

def sum(id, value):
    operation(id, value)

def start_minimum(item, property):
    return start_operation(item, property, operation_type.min)

def min(id, value):
    operation(id, value)

def start_maximum(item, property):
    return start_operation(item, property, operation_type.max)

def max(id, value):
    operation(id, value)

def start_average(item, property):
    return start_operation(item, property, operation_type.avg)

def avg(id, value):
    operation(id, value)

def start_operation(item, property, operation):
    if voxelfarm_framework:
        return voxelfarm_framework.start_operation(item, property, operation)
    return -1

def operation(id, value):
    if voxelfarm_framework:
        voxelfarm_framework.operation(id, value)

def get_material(component, voxel):
    if voxelfarm_framework:
        return voxelfarm_framework.get_material(component, voxel)
    return None

def get_channels(component):
    if voxelfarm_framework:
        count = voxelfarm_framework.get_channel_count(component)
        return range(0, count)
    return range(0)

def get_volume(component, voxel):
    if voxelfarm_framework:
        return voxelfarm_framework.get_volume(component, voxel)
    return 0

def get_channel_volume(component, voxel, channel):
    if voxelfarm_framework:
        return voxelfarm_framework.get_volume_channel(component, voxel, channel)
    return 0

def get_attribute_value(component, attribute, voxel):
    if voxelfarm_framework:
        return voxelfarm_framework.get_attribute(component, attribute, voxel)
    return 0

def get_channel_attribute_value(component, attribute, voxel, channel):
    if voxelfarm_framework:
        return voxelfarm_framework.get_attribute_channel(component, attribute, voxel, channel)
    return 0

def load_voxels(entity, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.load_voxels(entity, attributes, serialize_attributes(custom_attributes))
    return -1

def voxels_intersection(componentA, componentB, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.voxels_intersection(componentA, componentB, attributes, serialize_attributes(custom_attributes))
    return -1

def voxels_union(componentA, componentB, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.voxels_union(componentA, componentB, attributes, serialize_attributes(custom_attributes))
    return -1

def voxels_complement(componentA, componentB, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.voxels_complement(componentA, componentB, attributes, serialize_attributes(custom_attributes))
    return -1

def load_mesh(entity, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.load_mesh(entity, attributes, serialize_attributes(custom_attributes))
    return -1

def mesh_intersection(componentA, componentB, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.mesh_intersection(componentA, componentB, attributes, serialize_attributes(custom_attributes))
    return -1

def mesh_union(componentA, componentB, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.mesh_union(componentA, componentB, attributes, serialize_attributes(custom_attributes))
    return -1

def mesh_complement(componentA, componentB, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.mesh_complement(componentA, componentB, attributes, serialize_attributes(custom_attributes))
    return -1

def load_points(entity, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.load_points(entity, attributes, serialize_attributes(custom_attributes))
    return -1

def load_blocktree(entity, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.load_blocktree(entity, attributes, serialize_attributes(custom_attributes))
    return -1

def load_density(entity, attributes, custom_attributes):
    if voxelfarm_framework:
        return voxelfarm_framework.load_density(entity, attributes, serialize_attributes(custom_attributes))
    return -1

def load_texture(entity, type):
    if voxelfarm_framework:
        return voxelfarm_framework.load_texture(entity, type)
    return -1

def load_texture_stack():
    if voxelfarm_framework:
        return voxelfarm_framework.load_texture_stack()
    return ''

def apply_texture(component, texture, type):
    if voxelfarm_framework:
        return voxelfarm_framework.apply_texture(component, texture, type)

def apply_texture_stack(component, textures, type):
    if voxelfarm_framework:
        return voxelfarm_framework.apply_texture_stack(component, textures, type)

def new_region_scope(region):
    if voxelfarm_framework:
        return range(0, voxelfarm_framework.new_region_scope(region))
    return range(0)        

def render(component, material):
    if voxelfarm_framework:
        materialData = material.serialize() if material != None else ""
        voxelfarm_framework.render(component, materialData)

def log(message):
    if voxelfarm_framework:
        voxelfarm_framework.log(message)
    else:
        print(message)    

def new_material():
    return Material()

def input_duplicate(entity):
    if voxelfarm_framework:
        return voxelfarm_framework.input_duplicate(entity)
    return -1

def new_plane(a, b, c, d):
    if voxelfarm_framework:
        return voxelfarm_framework.new_plane(a, b, c, d)
    return -1

def get_enum_attribute_name_count(component, attribute_name):
    if voxelfarm_framework:
        return voxelfarm_framework.get_enum_attribute_count(component, attribute_name)
    return 0

def get_enum_attribute_name_value(component, attribute_name, index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_enum_attribute_name_value(component, attribute_name, index)
    return ''
    
def get_enum_attribute_count(component, attribute):
    if voxelfarm_framework:
        return voxelfarm_framework.get_enum_attribute_count(component, attribute)
    return 0

def get_enum_attribute_value(component, attribute, index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_enum_attribute_value(component, attribute, index)
    return ''
    
def set_component_attribute_merge_mode(component, attribute, merge_mode):
    if voxelfarm_framework:
        return voxelfarm_framework.set_component_attribute_merge_mode(component, attribute, merge_mode)

def valid_volume_channel(component, channel):
    if voxelfarm_framework:
        return voxelfarm_framework.valid_volume_buffer(component, channel)
    return False

def valid_attribute_channel(component, attribute, channel):
    if voxelfarm_framework:
        return voxelfarm_framework.valid_attribute_buffer(component, attribute, channel)
    return False

def valid_volume_buffer(component):
    if voxelfarm_framework:
        return voxelfarm_framework.valid_volume_buffer(component)
    return False

def valid_attribute_buffer(component, attribute):
    if voxelfarm_framework:
        return voxelfarm_framework.valid_attribute_buffer(component, attribute)
    return False

def get_voxel_volume():
    if voxelfarm_framework:
        return voxelfarm_framework.get_voxel_volume()
    return False

def load_pandas_dataframe(component):
    return None

def sum_pandas_result(name, df):
    pass

def export(component, name):
    if voxelfarm_framework:
        voxelfarm_framework.export(component, name)

def store(component):
    if voxelfarm_framework:
        voxelfarm_framework.store(component)

def add_property(layer, key, value):
    if voxelfarm_framework:
        voxelfarm_framework.add_property(layer, key, value)

# Generator

def init():
    pass

def set_vector(index, vx, vy, vz):
    if voxelfarm_framework:
        voxelfarm_framework.set_vector(index, vx, vy, vz)

def set_material(index, id):
    if voxelfarm_framework:
        voxelfarm_framework.set_material(index, id)

def set_field(index, value):
    if voxelfarm_framework:
        voxelfarm_framework.set_field(index, value)

def get_voxel_origin(index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_voxel_origin(index, 44)
    return (0,0,0)

def get_field_origin(index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_voxel_origin(index, 45)
    return (0,0,0)

def set_entity_bounds_x(v0, v1):
    if voxelfarm_framework:
        if voxelfarm_framework.set_entity_bounds(0, v0, v1) == 0:
            cellIsEmpty()

def set_entity_bounds_y(v0, v1):
    if voxelfarm_framework:
        if voxelfarm_framework.set_entity_bounds(1, v0, v1) == 0:
            cellIsEmpty()

def set_entity_bounds_z(v0, v1):
    if voxelfarm_framework:
        if voxelfarm_framework.set_entity_bounds(2, v0, v1) == 0:
            cellIsEmpty()

def get_field_origin_x(index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_voxel_origin_axis(index, 45, 0)
    return 0.0

def get_field_origin_y(index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_voxel_origin_axis(index, 45, 1)
    return 0.0

def get_field_origin_z(index):
    if voxelfarm_framework:
        return voxelfarm_framework.get_voxel_origin_axis(index, 45, 2)
    return 0.0

def get_voxel_origin_axis(index, dimension, axis):
    return (0,0,0)

def cellIsEmpty():
    pass

def test():
    pass




import zipfile
import json
import base64
import time
import requests
from datetime import datetime, timedelta
import os
import random
import pandas as pd
from concurrent.futures import ThreadPoolExecutor
from concurrent.futures import as_completed
from io import StringIO
import io
import sys
import voxelfarm

class mock_workflow_framework:

    def __init__(self, project, vf_api):  
        self.project = project
        self.vf_api = vf_api

    def input_string(self, id, label, default = ""):
        if id == 'project_id':
            return self.project
        elif id == 'project':
            return self.project
        
        return default

    def log(self, message):
        print(message)

    def progress(self, progress, message):
        print(str(progress) + ' ' + message)

    def get_entity(self, id = None):
        return self.vf_api.get_entity(id, self.project)

    def download_entity_file(self, filename, id = None):
        return ""
    
    def get_scrap_folder(self):
        return ""

    def get_tools_folder(self):
        return ""

    def get_entity_folder(self, id = None):
        return ""

    def download_entity_files(self, id = None):
        return ""

    def download_entity_file(self, filename, id = None):
        return ""

    def attach_file(self, filename, id = None):
        pass

    def attach_folder(self, folder, id = None):
        pass

    def upload(self, filename, name, id = None):
        pass

    def set_exit_code(self, code):
        print('Exit code: ' + str(code))

    def get_entity_file_list(self, id = None):
        return []
    
    def update_type(self):
        return None

    def stage_done(self):
        return False

    def get_vf_api(self):
        return self.vf_api
    
    def increment_counter(self, counter, offset = 1):
        return 0
    
    def decrement_counter(self, counter, offset = 1):
        return 0

    def set_counter(self, counter, value = 0):
        pass

    def get_working_dir(self):
        return ''

    def get_notify_emails(self):
        return ''

    def get_notify_users_url(self):
        return ''

    def get_workflow_api(self):
        return ''

    def publish_feed(self, filename):
        return {'success': True}

class rest:

    OrgId = '2343243456678890'
    class entity_type:
        View = 'VIEW'
        Project = 'PROJECT'
        VoxelTerrain = 'VOXSURF'
        RealtimeVoxelTerrain = 'VOXOP'
        BlockModel = 'VOXBM'
        IndexedPointCloud = 'IDXPC'
        VoxelGenerator = 'VOXGEN'
        RawPointCloud = 'RAWPC'
        RawHeightmap = 'RAWHM'
        RawBlockModel = 'RAWBM'
        RawMesh = 'RAWMESH'
        IndexedMesh = 'IDXMESH'
        VoxelMesh = 'VOXMESH'
        OrthoImagery = 'ORTHO'
        IndexedOrthoImagery = 'IDXORTHO'
        Program = 'PROGRAM'
        Folder = 'FOLDER'
        RawDensity = 'RAWDENSITY'
        IndexedDensity = 'IDXDENSITY'
        MaterialTracking = 'VOXMT'
        MaterialTrackingOperation = 'VOXMTOP'
        RawDrillHoles = 'RAWDH'
        DrillHoles = 'IDXDH'
        OrthoVox='ORTHOVOX'
        Process='PROCESS'
        Export='EXPORT'
        Report = 'REPORT'
        VoxelPC = 'VOXPC'
        RawGeoChem = 'RAWGEOCHEM'
        GeoChem = 'GEOCHEM'

    class lambda_type:
        Generator = 'VOXEL'
        Report = 'REPORT'
        View = 'VIEW'
        Workflow = 'WORKFLOW'
        Process = 'PROCESS'

    class process_lambda_type:
        Workflow = 'WORKFLOW'
        Process = 'PROCESS'

    CRSFields = [
        "coord_origin_x",
        "coord_origin_y",
        "coord_origin_z",
        "coord_hdatum",
        "coord_vdatum",
        "coord_projection",
        "coord_unit",
        "voxel_size",
        "coord_projection_tm_falseEasting",
        "coord_projection_tm_falseNorthing",
        "coord_projection_tm_latOriginDeg",
        "coord_projection_tm_longMeridianDeg",
        "coord_projection_tm_scaleFactor",
        "coord_projection_lcc_falseEasting",
        "coord_projection_lcc_falseNorthing",
        "coord_projection_lcc_latOfOriginDeg",
        "coord_projection_lcc_longOfOriginDeg", 
        "coord_projection_lcc_firstStdParallelDeg",
        "coord_projection_lcc_secondStdParallelDeg",
        "coord_projection_aeac_falseEasting",
        "coord_projection_aeac_falseNorthing",
        "coord_projection_aeac_latOfOriginDeg",
        "coord_projection_aeac_longOfOriginDeg", 
        "coord_projection_aeac_firstStdParallelDeg",
        "coord_projection_aeac_secondStdParallelDeg",
        "coord_projection_amg_zone",
        "coord_projection_utm_easting",
        "coord_projection_utm_northing"]

    DefaultFields = {
        entity_type.VoxelTerrain : {
            'include_classification' : '0 1 2 3 4 5 6 7 8 9 10 11 12 13 14 15 16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31',
        },
        entity_type.VoxelMesh : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ',
        },
        entity_type.IndexedMesh : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ',
        },
        entity_type.BlockModel : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ',
            'bm_type': 'VOXBM1'
        },
        entity_type.OrthoVox : {
            'translate_x' : '0',
            'translate_y' : '0',
            'translate_z' : '0',
            'scale_x' : '1',
            'scale_y' : '1',
            'scale_z' : '1',
            'rotate_x' : '0',
            'rotate_y' : '0',
            'rotate_z' : '0',
            'rotation_order' : 'ZXZ'
        }
    }

    HTTPSession = requests.Session()

    def __init__(self, apiurl="https://api.voxelfarm.com"):
        self.proxy = None
        self.VFAPIURL = apiurl
        self.VFAPIURL_ENTITY = self.VFAPIURL + '/entity.ashx'
        self.VFAPIURL_FILE = self.VFAPIURL + '/file.ashx'
        self.VFAPIURL_EVENT = self.VFAPIURL + '/events.ashx'
        self.token: str = None
        self.aad_credentials : dict = None
        self.ts_api_version = "2020-07-31"
        self.ts_environmentName = os.getenv('vf_ts_hub_name')
        self.ts_applicationName = os.getenv('vf_ts_app_name')

    def set_proxy(self, proxy):
        self.proxy = proxy

    class api_result:
        def __init__(self):      
            self.success = False
            self.error_info = ''

    class creation_result(api_result):
        def __init__(self):      
            self.id = ''

    class multiple_creation_result():
        def __init__(self):      
            self.success = False
            self.error_info = ''
            self.ids = []

    class wait_result(api_result):
        def __init__(self):      
            self.complete = False
            self.times = {}

    def add_default_fields(self, type, fields):
        if type in self.DefaultFields:
            for field in self.DefaultFields[type]:
                if not field in fields:
                    fields[field] = self.DefaultFields[type][field]
        if not 'file_date' in fields:
            fields['file_date'] = str(1000 * int(time.time()))

    class crs_result(api_result):
        def __init__(self):      
            self.crs = {}

    def get_project_crs(self, project):
        result = self.crs_result()
        params = {'id': project, 'project': project}

        projectRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if projectRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {projectRequest.status_code}' 
            return result
        entityJson = json.loads(projectRequest.text)
        values = {
            'coord_hunit': entityJson['coord_unit'],
            'coord_vunit': entityJson['coord_unit']
        }
        for field in self.CRSFields:
            if field in entityJson:
                values[field] = entityJson[field]
            else:
                values[field] = '0'
        result.crs = values
        result.success = True
        return result

    def get_entity(self, id, project = None):
        params={'id': id}
        if project:
            params['project'] = project
        projectRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if projectRequest.status_code != 200:
            return None
        entityJson = json.loads(projectRequest.text)
        return entityJson

    def get_projects_names(self):
        result = self.collection_result()
        entityRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, headers=self.get_auth_headers())
        if (entityRequest.status_code == 200):
            entityJson = entityRequest.json()
            items = []
            for x in entityJson:
                v = entityJson[x]
                if v["type"] == "PROJECT_REF":
                    items.append(v["project_ref"])

            result.items = items
            result.success = True
        else:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'  
        return result

    def create_project(self, name, fields):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        self.add_default_fields(type, fields)
        data = {'operation': 'create', 'data': json.dumps(fields)}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, data=data, headers=self.get_auth_headers(), proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        result.id = apiResult['id']
        return result

    def delete_project(self, project):
        items = []
        # Get all entities in project
        result = self.get_collection(project)
        if not result.success:
            print(result.error_info)
            exit(3)
        entities = result.items
        for e in entities:
            entity = entities[e]
            if entity['type'] != 'USER':
                items.append(entity["ID"])
        # Get all user_ref objects for project (users associated with the project)
        result = self.get_collection('users:' + project)
        if not result.success:
            print(result.error_info)
            exit(4)
        user_refs = result.items
        for user_ref in user_refs:
            if user_refs[user_ref]['type'] == 'USER_REF':
                user_id = user_refs[user_ref]['user_ref']
                items.append(user_id + ':' + project)
                items.append(project + ':' + user_id)
        items.append('R3Vlc3Q=:' + project) # guest user
        self.delete_entities(project, items)

    def clean_project(self, project):
        items = []
        # Get all entities in project
        result = self.get_collection(project)
        if not result.success:
            print(result.error_info)
            exit(3)
        entities = result.items
        for e in entities:
            entity = entities[e]
            if entity['type'] != 'USER' and entity['type'] != 'PROJECT':
                items.append(entity["ID"])

        self.delete_entities(project, items)        

    def create_entity_raw(self, project, type, name, fields, crs):
        result = self.creation_result()
        if fields == None:
            fields = {}
            
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = type

        if 'state' not in fields:
            fields['state'] = 'COMPLETE'

        fields['project'] = project
        for key, value in crs.items():
            fields['entity_' + key] = value
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}' 
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        result.id = apiResult['id']
        return result

    def update_entity(self, project, id, fields):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['ID'] = id
        fields['project'] = project
        params={'project': project, 'id': id}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.success = True
        result.id = id
        return result

    def create_folder(self, project, folder, name, fields = {}):
        result = self.creation_result()
        if folder == None:
            folder = '0'

        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'FOLDER'
        fields['state'] = 'COMPLETE'
        fields['project'] = project
        fields['file_folder'] = folder
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.success = True
        result.id = apiResult['id']
        return result

    def create_entity_processed(self, project, type, name, fields, crs, callback=None, update_type = None):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = type
        fields['project'] = project
        fields['state'] = 'PARTIAL'
        if update_type:
            fields['callback_update_type'] = update_type
        for key, value in crs.items():
            fields[key] = value

        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'PROCESS'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        return result

    def create_process_entity(self, project, name, code, fields, inputs, files = None, callback = None, update_type = None):
        result = self.creation_result()

        # Create the Entity
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'PROCESS'
        fields['state'] = 'PARTIAL'
        fields['project'] = project

        fields['code'] = code

        python_file = False
        code_origin = None
        if files:
            for file in files:
                if code in file:
                    python_file = True
                    break

        if python_file:
            code_origin = 'PYTHON_FILE'
        else:
            program_entity = self.get_entity(code)
            if program_entity:
                code_origin = 'PROGRAM'

        if code_origin:
            fields['code_origin'] = code_origin
        else:
            result.success = False
            result.error_info = f'We cannot detect the code_origin from this code {code}'
            return result

        if update_type:
            fields['callback_update_type'] = update_type

        if inputs:
            for input in inputs:
                fields['input_value_' + input] = inputs[input]
                fields['input_filter_' + input] = '0'
                fields['input_type_' + input] = '0'
                fields['input_label_' + input] = '0'

        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']

        # Attach files to the entity
        if files:
            for file in files:
                if (not os.path.exists(file)):
                    result.success = False
                    result.error_info = 'File not found: ' + file
                    return result

                attach_files = {'file': open(file, 'rb')}
                apiResult = self.attach_files(project=project, id=result.id, files=attach_files)
                if not apiResult.success:
                    result.success = False
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                    return result

        fields.clear()
        fields['partial_status'] = ""
        fields['partial_progress'] = "0"

        result = self.update_entity(
            project=project, 
            id=result.id, 
            fields=fields)

        # Process the Entity
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'PROCESS'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        return result        

    def create_workflow_entity(self, project, name, fields, inputs):
        result = self.creation_result()

        # Create the Entity
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'PROCESS'
        fields['process_type'] = 'WORKFLOW'
        fields['state'] = 'PARTIAL'
        fields['background_process'] = '1'
        fields['project'] = project

        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'

        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']

        fields.clear()
        fields['partial_status'] = ""
        fields['partial_progress'] = "0"

        result = self.update_entity(
            project=project, 
            id=result.id, 
            fields=fields)

        # Process the Entity
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'PROCESS'}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        return result        

    def create_report(self, project, program, lod, region, name, fields, inputs, callback=None, update_type = None):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'REPORT'
        fields['program'] = program
        fields['project'] = project
        fields['region'] = region
        fields['lod'] = str(lod)
        fields['state'] = 'PARTIAL'
        if update_type:
            fields['callback_update_type'] = update_type
        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result

        result.id = apiResult['id']
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'RUN_REPORT'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
        else:
            result.success = True

        return result

    def create_export(self, project, program, lod, region, name, fields, inputs, callback=None):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'EXPORT'
        fields['program'] = program
        fields['project'] = project
        fields['region'] = region
        fields['lod'] = str(lod)
        fields['state'] = 'PARTIAL'

        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.id = apiResult['id']
        params={'project': project, 'org': self.OrgId, 'id': result.id}
        data={'process':'RUN_EXPORT'}
        if callback != None:
            data['callback'] = callback
        entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
        else:
            result.success = True

        return result

    def create_view(self, project, name, view_type, inputs, fields):
        input_fields = {
            'file_folder' : '0',
            'view_type' : view_type,
            'virtual' : '1',
            'state' : 'COMPLETE',
            'color_legend_attribute' : '',
            'color_legend_attribute_index' : '-1',
            'color_legend_gradient' : 'isoluminant_cgo_70_c39_n256',
            'color_legend_interpolate_gradient' : '1',
            'color_legend_mode' : '2',
            'color_legend_range_max' : '100',
            'color_legend_range_min' : '0',
            'color_legend_range_step' : '1',
            'color_legend_reverse_gradient' : '0',
            'file_date' : str(1000 * int(time.time())),
            'file_type' : 'VIEW',
            'input_filter_colors' : '0',
            'input_filter_e' : '8',
            'input_filter_normals' : '0',
            'input_label_colors' : 'Use Ortho-imagery',
            'input_label_e' : 'Terrain',
            'input_label_normals' : 'Use high resolution detail',
            'input_type_colorlegend' : '7',
            'input_type_colors' : '6',
            'input_type_e' : '3',
            'input_type_normals' : '6',
            'input_value_colors' : '0',
            'input_value_normals' : '0',         
        }
            
        if inputs:    
            for key in inputs:
                input_fields['input_value_' + key] = inputs[key]

        if fields:
            for key in fields:
                input_fields[key] = fields[key]
            
        api_result = self.create_entity_raw(
            project=project,
            type=self.entity_type.View,
            name=name,
            fields=input_fields,
            crs={}
        )

        if not api_result.success:
            result = self.creation_result()
            result.error_info = api_result.error_info
            return result
        
        view_object = api_result.id
        
        result = self.create_entity_raw(
            project=project,
            type=self.entity_type.View,
            name=name,
            fields={
                'file_folder' : folder,
                'view_type' : 'container',
                'state' : 'COMPLETE',
                'entity_container' : view_object
            },
            crs={}
        )

        return result

    def reprocess_entity(self, id):
        result = self.creation_result()
        result.id = id

        report_entity = self.get_entity(id)
        if report_entity:
            project = report_entity["project"]

            params = {
                'project': project,
                'id' : id
            }

            fields = {
                'state': 'PARTIAL',
                'partial_progress': '0',
                'partial_status': '',
                'proccessing_time': '0'
            }

            entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
            if entityRequest.status_code != 200:
                result.success = False
                result.error_info = f'HTTP Error code {entityRequest.status_code}'
                return result

            apiResult = json.loads(entityRequest.text)
            if ('result' not in apiResult) or (apiResult['result'] != 'success'):
                result.success = False
                if 'error_info' in apiResult:
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                else:    
                    result.error_info = 'error_info not found'
                return result

            params={'project': project, 'org': self.OrgId, 'id': id}
            data={'process':'RUN_REPORT'}
            entityRequest = self.HTTPSession.post(self.VFAPIURL_EVENT, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
            if entityRequest.status_code != 200:
                result.success = False
                result.error_info = f'HTTP Error code {entityRequest.status_code}'
                return result
            apiResult = json.loads(entityRequest.text)
            if ('result' not in apiResult) or (apiResult['result'] != 'success'):
                result.success = False
                if 'error_info' in apiResult:
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                else:    
                    result.error_info = 'error_info not found'
            else:
                result.success = True
        else:
            result.success = False

        return result        

    def create_lambda_python(self, project, type, name, code, fields, files=None):
        result = self.creation_result()
        if fields == None:
            fields = {}

        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'PROGRAM'
        fields['program_type'] = type
        fields['state'] = 'COMPLETE'

        if code.endswith('.py'):
            fields['code_origin'] = 'PYTHON_FILE'
            fields['code'] = code
        else:
            generatorLambdaCode = code.encode('ascii')
            generatorLambdaCode = base64.b64encode(generatorLambdaCode).decode('ascii')
            fields['code'] = generatorLambdaCode

        fields['project'] = project
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = requests.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        
        lambda_id = apiResult['id']
        if files:
            for file in files:
                if (not os.path.exists(file)):
                    result.success = False
                    result.error_info = 'File not found: ' + file
                    return result

                attach_files = {'file': open(file, 'rb')}
                apiResult = self.attach_files(project=project, id=lambda_id, files=attach_files)
                if not apiResult.success:
                    result.success = False
                    result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
                    return result
                        
        result.success = True
        result.id = lambda_id
        return result
    
    def create_generator(self, project, program, name, fields, inputs):
        result = self.creation_result()
        if fields == None:
            fields = {}
        fields['name'] = name
        fields['type'] = 'FILE'
        fields['file_type'] = 'VOXGEN'
        fields['program'] = program
        fields['state'] = 'COMPLETE'
        fields['project'] = project
        for input in inputs:
            fields['input_value_' + input] = inputs[input]
            fields['input_filter_' + input] = '0'
            fields['input_type_' + input] = '0'
            fields['input_label_' + input] = '0'
        self.add_default_fields(type, fields)
        params={'project': project}
        entityRequest = requests.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), json=fields, proxies=self.proxy)
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        result.success = True
        result.id = apiResult['id']
        return result

    def is_processing_complete(self, project, ids):
        result = self.wait_result()
        for id in ids:
            params = {'id': id, 'project': project}
            entityRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers())
            if entityRequest.status_code == 200:
                entityJson = json.loads(entityRequest.text)
                if not 'state' in entityJson:
                    result.complete = False
                    result.success = False
                    result.error_info = json.dumps(entityJson)
                    break
                if entityJson['state'] == 'PARTIAL':
                    result.complete = False
                    result.success = True
                    break
                if entityJson['state'] == 'COMPLETE':
                    result.complete = True
                    result.success = True
                    if 'processing_time' in entityJson:
                        result.times[id] = entityJson['processing_time']
                    else:    
                        result.times[id] = ''
                if entityJson['state'] == 'ERROR':
                    result.complete = False
                    result.success = False
                    result.error_info = base64.b64decode(entityJson['error_info']).decode('ascii')
                    break
            else:
                result.complete = False
                result.success = False
                result.error_info = 'HTTP Error ' + str(entityRequest.status_code)
                break

        return result

    def delete_entities(self, project, ids):
        result = self.api_result()
        deleteIdentifierList = ''
        for id in ids:
            deleteIdentifierList = deleteIdentifierList + id + ' '
        params = {'project': project, 'org': self.OrgId, 'operation': 'delete'}
        data = {'operation': 'delete', 'items': deleteIdentifierList}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers(), data=data, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = 'Error deleting entities'
            return result

        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        else:    
            result.success = True
        return result

    def attach_files(self, project, id, files):
        result = self.api_result()
        params = {'project': project, 'org': self.OrgId, 'id': id}
        entityRequest = self.HTTPSession.post(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), files=files, proxies=self.proxy)
        if entityRequest.status_code != 200:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
            return result
        apiResult = json.loads(entityRequest.text)
        if ('result' not in apiResult) or (apiResult['result'] != 'success'):
            result.success = False
            if 'error_info' in apiResult:
                result.error_info = base64.b64decode(apiResult['error_info']).decode('ascii')
            else:    
                result.error_info = 'error_info not found'
            return result
        else:    
            result.success = True
        return result

    def get_file(self, project, entity, filename):
        params = {'id': entity, 'project': project, 'org': self.OrgId, 'filename': filename}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if entityRequest.status_code != 200:
            return ''
        return entityRequest.text

    def get_binary_file(self, project, entity, filename):
        params = {'id': entity, 'project': project, 'org': self.OrgId, 'filename': filename}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if entityRequest.status_code != 200:
            return None
        return entityRequest.content  

    def download_file(self, project, entity, filename, destination):
        params = {'id': entity, 'project': project, 'org': self.OrgId, 'filename': filename}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)
        if (entityRequest.status_code) != 200:
            return False
        with open(destination, 'wb') as file:
            file.write(entityRequest.content)
        return True

    def get_file_list(self, project, entity):
        params = {'id': entity, 'project': project, 'org': self.OrgId}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_FILE, params=params, headers=self.get_auth_headers(), proxies=self.proxy)

        list = []
        if entityRequest.status_code == 200:
            for line in entityRequest.text.split("\n"):
                if not line.strip():
                    continue
                list.append(line.lstrip())        
        return list

    class collection_result(api_result):
        items = None

    def get_collection(self, id):
        result = self.collection_result()
        params = {'project': id}
        entityRequest = self.HTTPSession.get(self.VFAPIURL_ENTITY, params=params, headers=self.get_auth_headers())
        if (entityRequest.status_code == 200):
            entityJson = json.loads(entityRequest.text)
            result.items = entityJson
            result.success = True
        else:
            result.success = False
            result.error_info = f'HTTP Error code {entityRequest.status_code}'
        return result

    def compute_region_from_aabbs(self, entity_id):
        entity = self.get_entity(entity_id)
        min_x = float(entity["aabb_min_x"]) - 20
        max_x = float(entity["aabb_max_x"]) + 20
        min_y = float(entity["aabb_min_y"]) - 20
        max_y = float(entity["aabb_max_y"]) + 20
        min_z = float(entity["aabb_min_z"]) - 20
        max_z = float(entity["aabb_max_z"]) + 20
        region = '1'
        region += "," + str(min_z)
        region += "," + str(max_z)
        region += "," + str(min_x)
        region += "," + str(min_y)
        region += "," + str(max_x)
        region += "," + str(min_y)
        region += "," + str(max_x)
        region += "," + str(max_y)
        region += "," + str(min_x)
        region += "," + str(max_y)
        return region
    
    def get_dataframe(self, project_id, entity_id, file_name):
        file = self.get_file(project_id, entity_id, 'report.csv')
        if file:
            content = StringIO(file)
            return pd.read_csv(content)

        return pd.DataFrame()

    def deploy_workflow(self, project, code_path, code):
        print('Deploying Workflow')

        print(f'Fetch project entity: {project}')
        project_entity = self.get_entity(project, project)

        print(f'Compress code files')
        code_file = os.path.join(code_path, 'code.zip')

        def zipdir(path, ziph):
            for root, dirs, files in os.walk(path):
                if '__pycache__' not in root:
                    for file in files:
                        if file != 'code.zip':
                            p = os.path.relpath(os.path.join(root, file), path)
                            ziph.write(os.path.join(root, file), p)

        with zipfile.ZipFile(code_file, 'w', zipfile.ZIP_DEFLATED) as zipf:
            zipdir(code_path, zipf)

        create_workflow = True
        if 'workflow_id' in project_entity:
            print(f'Workflow already initialized')
            print(f'Check if entity exists {project_entity["workflow_id"]}')
            if self.get_entity(project_entity['workflow_id'], project):
                print(f'Update entity code')

                workflow_id = project_entity['workflow_id']
                create_workflow = False

                fields = {
                    'code' : code
                }
                result = self.update_entity(
                    project=project, 
                    id=workflow_id, 
                    fields=fields)

                #attach code  
                workflow_code_file = open(code_file, 'rb')
                files = {'file': workflow_code_file}
                result = self.attach_files(
                    project=project,
                    id=workflow_id,
                    files=files)
                
                workflow_code_file.close()

                if not result.success:
                    print('Error uploading workflow code ' + result.error_info)
                    exit()
            else:
                print(f'Entity {project_entity["workflow_id"]} does not exist. Check your .env variables')
                exit()

        if create_workflow:
            print(f'Create program entity')
            result = self.create_lambda_python(
                    project=project, 
                    type=voxelfarm.lambda_type.Workflow,
                    name='workflow', 
                    code=code, 
                    fields={
                        'virtual': '1',
                        'file_folder': '0'
                    },
                    files=[code_file])

            if not result.success:
                print('Error creating workflow folder.')
                exit()

            workflow_id = result.id
            print(f'Program entity {workflow_id} created')

            fields = {
                'workflow_id' : workflow_id
            }

            print(f'Saving changes to project {project}...')

            result = self.update_entity(
                project=project, 
                id=project, 
                fields=fields)

            if result.success:
                print('Saved changes to project')
            else:
                print(f'Error saving project: {result.error_info}')

        print('Code file uploaded.')
        os.remove(code_file)

        voxelfarm.voxelfarm_framework = mock_workflow_framework(project, self)

        sys.path.append(code_path)
        code_file = os.path.join(code_path, code)
        with open(code_file, "r") as archivo:
            codigo = archivo.read()
            exec(codigo)

        print('Done.')

    def workflow_get_product_property(self, project_id, product_id, property, version = 0):
        project = self.get_entity(project_id, project_id)
        if project:
            product_key = 'workflow_folder_' + product_id
            if product_key in project:
                product_folder_id = project[product_key]
                if version > 0:
                    apiResult = self.get_collection(project_id)
                    if apiResult.success:
                        version_name = 'Version ' + str(version)
                        entities = apiResult.items
                        for e in entities:
                            entity = entities[e]
                            if ('file_folder' in entity) and (entity['file_folder'] == product_folder_id) and ('file_type' in entity) and (entity['file_type'] == 'FOLDER'):
                                if entity["name"] == version_name:
                                    extended_prop = 'version_' + property
                                    if extended_prop in entity:
                                        return entity[extended_prop]
                                    break    
                else:
                    product_entity = self.get_entity(product_folder_id, project_id)
                    if product_entity and ('version_active' in product_entity):
                        active_version_id = product_entity['version_active']
                        active_version = self.get_entity(active_version_id, project_id)
                        if active_version:
                            extended_prop = 'version_' + property
                            if extended_prop in active_version:
                                return active_version[extended_prop]
        return None

    def workflow_get_parameter_dataframe(self, project_id, product_id, version = 0):
        attribute_product = self.workflow_get_product_property(project_id, product_id, 'report_entity', version)
        if attribute_product:
            attribute = self.get_file(project_id, attribute_product, 'report.csv')
            if attribute:
                content = StringIO(attribute)
                return pd.read_csv(content)
        return pd.DataFrame()

    def workflow_get_entity_id(self, project_id, product_id, alias_id, version = 0):
        project = self.get_entity(project_id, project_id)
        if project:
            product_key = 'workflow_folder_' + product_id
            if product_key in project:
                product_folder_id = project[product_key]
                if version > 0:
                    apiResult = self.get_collection(project_id)
                    if apiResult.success:
                        version_name = 'Version ' + str(version)
                        entities = apiResult.items
                        for e in entities:
                            entity = entities[e]
                            if ('file_folder' in entity) and (entity['file_folder'] == product_folder_id) and ('file_type' in entity) and (entity['file_type'] == 'FOLDER'):
                                if entity["name"] == version_name:
                                    extended_prop = 'alias_' + alias_id
                                    if extended_prop in entity:
                                        return entity[extended_prop]
                                    break    
                else:
                    product_entity = self.get_entity(product_folder_id, project_id)
                    if product_entity and ('version_active' in product_entity):
                        active_version_id = product_entity['version_active']
                        active_version = self.get_entity(active_version_id, project_id)
                        if active_version:
                            extended_prop = 'alias_' + alias_id
                            if extended_prop in active_version:
                                return active_version[extended_prop]
        return None

    def workflow_get_dataframe(self, project_id, product_id, alias_id, file_name = None, version = 0):
        attribute_product = self.workflow_get_entity_id(project_id, product_id, alias_id, version)
        if attribute_product:
            dataframe_file = 'report.csv'
            if file_name:
                dataframe_file = file_name

            split_tup = os.path.splitext(dataframe_file)
            file_extension = split_tup[1].lower()
            if file_extension == '.csv':
                attribute = self.get_file(project_id, attribute_product, dataframe_file)
                if attribute:
                    content = StringIO(attribute)
                    return pd.read_csv(content)
            elif file_extension == '.ftr':
                attribute = self.get_binary_file(project_id, attribute_product, dataframe_file)
                if attribute:
                    content = io.BytesIO(attribute)
                    return pd.read_feather(content)

        return pd.DataFrame()

    def workflow_get_timings(self, project_id):
        entities = []
        collections = self.get_collection(project_id).items
        
        for entity_id in collections:
            entity = collections[entity_id]
            if "processing_start_date" in entity and "file_folder" in entity and entity["file_folder"] in collections:
                file_folder_entity = collections[entity["file_folder"]]
                processing_time = float(entity["processing_time"]) if "processing_time" in entity else 0
                if processing_time > 0 and "version_number" in file_folder_entity:
                    file_date = int(entity["file_date"]) if "file_date" in entity else 0
                    if file_date > 0:
                        processing_start_date = int(entity["processing_start_date"])
                    queue_time = processing_start_date - file_date
                    workflow_time = file_date - int(file_folder_entity["file_date"]) if "file_date" in file_folder_entity and int(file_folder_entity["file_date"]) > 0 else 0
                    entities.append({
                            "id": entity["ID"],
                            "file_folder": entity["file_folder"],
                        "file_date": pd.to_datetime(file_date / 1000, unit='s'),
                        "workflow_name": file_folder_entity["workflow_product"] if "workflow_product" in file_folder_entity else "N/A",
                            "name": entity["name"],
                            "state": entity["state"],
                            "version": int(file_folder_entity["version_number"]),
                        "workflow_time_ms": workflow_time,
                        "processing_time_ms": processing_time,
                        "queue_time_ms": queue_time
                    })

        return pd.DataFrame.from_dict(entities)    

#region AAD_Credentials
    def set_file_credentials(self, filename: str):
        self.aad_credentials = self.get_env_file_as_dict(filename)
        self.token = self.get_token()
        if (self.token == None):
            print("credentials file: "+filename )
            raise Exception("Check that the credentials are correct")

    def set_credentials(self, tenant: str, client_id: str, client_secret: str):
        self.aad_credentials = {
            "TENANT": tenant,
            "CLIENT_ID": client_id,
            "CLIENT_SECRET":client_secret
        }
        self.token = self.get_token()
        if (self.token == None):
            print("Set Credentials failed ClientId : "+ client_id)
            raise Exception("Check that the credentials are correct")

    def get_env_file_as_dict(self, path: str) -> dict:
        if not os.path.isfile(path):
            t_path = os.path.dirname(__file__)+ path
            if os.path.isfile(t_path):    
                path = t_path
        with open(path, 'r') as f:
            return dict(tuple(line.replace('\n', '').split('=')) for line 
                        in f.readlines() if not line.startswith('#'))

    def get_token(self, ressource=None):
        if ressource is None:
            ressource = self.aad_credentials["CLIENT_ID"]

        if (self.aad_credentials["TENANT"] != None):
            authority = "https://login.microsoftonline.com/"+self.aad_credentials["TENANT"]+"/oauth2/token"
            headers = {'Content-Type': 'application/x-www-form-urlencoded'}
            data = {
                "grant_type" : "client_credentials",
                "client_id" : self.aad_credentials["CLIENT_ID"],
                "client_secret": self.aad_credentials["CLIENT_SECRET"],
                "scope" : "https://graph.microsoft.com/.default",
                "resource": ressource,
            }

            HTTPSession = requests.Session()
            tokenRequest = HTTPSession.post(authority, data=data, headers=headers)
            HTTPSession.close()

            if tokenRequest.status_code == 200:
                tokenResult = tokenRequest.json()
                return tokenResult["access_token"]
        return None

    def check_ExpiredToken(self, token: str):
        body = token.split(".")[1]
        body += "=" * ((4 - len(body) % 4) % 4) #ugh
        txt = base64.b64decode(body).decode('utf-8')
        jwt = json.loads(txt)
        expire = int(jwt["exp"])
        return (datetime.fromtimestamp(expire)) < (datetime.now()+timedelta(minutes=2))

    def get_auth_headers(self):
        if (self.token==None):
            return None
        if self.check_ExpiredToken(self.token):
            self.token = self.get_token()    
        
        return { 'Authorization': f'Bearer {self.token}' }
# endregion AAD_Credentials

# region GetTimeSeriesData
    def get_time_series_data_by_search(self, timespan, projectID, spatialID=None, objectID=None, properties=None, filter=None):
        authorizationToken = self.get_token(
            ressource="https://api.timeseries.azure.com")
        searchString = ""
        if spatialID is not None and spatialID != "" and spatialID != "*":
            searchString = searchString + spatialID + ", "
        if objectID is not None and objectID != "" and objectID != "*":
            searchString = searchString + objectID

        if(searchString[len(searchString)-2:] == ", "):
            searchString = searchString[:len(searchString)-2]

        url = (
            "https://"
            + self.get_environment_id(authorizationToken)
            + ".env.timeseries.azure.com/timeseries/instances/search?"
        )
        querystring = {"api-version": self.ts_api_version}

        payload = {"searchString": f"\"{searchString}\"",
                   "instances": {"pageSize": 100}}

        headers = {
            "Authorization": f"Bearer {authorizationToken}",
            "Content-Type": "application/json",
            "cache-control": "no-cache",
        }

        try:
            jsonResponse = requests.request(
                "POST",
                url,
                data=json.dumps(payload),
                headers=headers,
                params=querystring,
            )
            jsonResponse.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            raise Exception("The request to the TSI api timed out.")
        except requests.exceptions.HTTPError:
            raise Exception(
                "The request to the TSI api returned an unsuccessfull status code.")

        response = json.loads(jsonResponse.text)
        if response["instances"]["hits"]:
            result = response["instances"]["hits"]
            processes = []
            with ThreadPoolExecutor(max_workers=100) as executor:
                for i, _ in enumerate(result):
                    processes.append(
                        executor.submit(
                            lambda p: self.get_data_by_id(*p),
                            [
                                result[i]["timeSeriesId"],
                                properties,
                                filter,
                                timespan,
                                authorizationToken,
                            ],
                        )
                    )

                while "continuationToken" in list(response["instances"].keys()):
                    headers["x-ms-continuation"] = response["instances"]["continuationToken"]
                    jsonResponse = requests.request(
                        "POST",
                        url,
                        data=json.dumps(payload),
                        headers=headers,
                        params=querystring,
                    )
                    jsonResponse.raise_for_status()

                    response = json.loads(jsonResponse.text)
                    result = response["instances"]["hits"]
                    for i, _ in enumerate(result):
                        processes.append(
                            executor.submit(
                                lambda p: self.get_data_by_id(*p),
                                [
                                    result[i]["timeSeriesId"],
                                    properties,
                                    filter,
                                    timespan,
                                    authorizationToken,
                                ],
                            )
                        )
            return self.build_dataframe_from_processes(processes, properties)
        else:
            return pd.DataFrame()

    def get_time_series_data_by_ids(self, projectID, timeseriesIDs, timespan):
        authorizationToken = self.get_token(
            ressource="https://api.timeseries.azure.com")

        processes = []
        with ThreadPoolExecutor(max_workers=100) as executor:
            for i, id in enumerate(timeseriesIDs):
                processes.append(
                    executor.submit(
                        lambda p: self.get_data_by_id(*p),
                        [
                            [id["spatialID"], id["objectID"]],
                            id["properties"] if "properties" in id else None,
                            id["filter"] if "filter" in id else None,
                            timespan,
                            authorizationToken,
                        ],
                    )
                )

        return self.build_dataframe_from_processes(processes, None)

    def get_environment_id(self, authorizationToken):
        url = "https://api.timeseries.azure.com/environments"

        querystring = {"api-version": self.ts_api_version}

        payload = ""
        headers = {
            "x-ms-client-application-name": self.ts_applicationName,
            "Authorization": f"Bearer {authorizationToken}",
            "Content-Type": "application/json",
            "cache-control": "no-cache",
        }

        try:
            response = requests.request(
                "GET",
                url,
                data=payload,
                headers=headers,
                params=querystring,
                timeout=10,
            )
            response.raise_for_status()
        except requests.exceptions.ConnectTimeout:
            raise Exception("The request to the TSI api timed out.")
        except requests.exceptions.HTTPError:
            raise Exception(
                "The request to the TSI api returned an unsuccessfull status code.")

        environments = json.loads(response.text)["environments"]
        environmentId = None
        for environment in environments:
            if environment["displayName"] == self.ts_environmentName:
                environmentId = environment["environmentId"]
                break
        if environmentId == None:
            raise Exception(
                "TSI environment not found. Check the spelling or create an environment in Azure TSI.")

        return environmentId

    def get_data_by_id(
        self,
        timeseries,
        properties,
        filter,
        timespan,
        authorizationToken,
    ):

        url = (
            "https://"
            + self.get_environment_id(authorizationToken)
            + ".env.timeseries.azure.com/timeseries/query?"
        )
        querystring = {"api-version": self.ts_api_version}
        requestType = "getEvents"

        result = {}

        projectedProperties = None
        if(properties != "*" and properties != None and properties != ["*"]):
            projectedProperties = properties
        retry = 0

        payload = {
            requestType: {
                "timeSeriesId": timeseries,
                "searchSpan": {"from": timespan[0], "to": timespan[1]},
                "take": 250000,
                "projectedProperties": projectedProperties,
            }
        }

        if filter is not None and filter != "":
            payload[requestType]["filter"] = {
                "tsx": filter
            }

        headers = {
            "x-ms-client-application-name": self.ts_applicationName,
            "Authorization": f"Bearer {authorizationToken}",
            "Content-Type": "application/json",
            "cache-control": "no-cache",
        }

        while retry < 50:  # Looping with a small random wait if the API returns us problems. This solve an issue where consicutive requests goes beyond TSI max number of concurrent requests.
            try:
                jsonResponse = requests.request(
                    "POST",
                    url,
                    data=json.dumps(payload),
                    headers=headers,
                    params=querystring,
                )
                jsonResponse.raise_for_status()
                break
            except requests.exceptions.HTTPError:
                error = json.loads(jsonResponse.text)
                if(error["error"]["code"] == "TooManyRequests"):
                    # Too many requests at the same time. Waiting for a bit before retrying."
                    time.sleep(random.randint(1, 5))
                    retry += 1
                    continue
                else:
                    raise Exception(
                        "The request to the TSI api returned an unsuccessfull status code.")

        response = json.loads(jsonResponse.text)
        result = response
        result["timeseries"] = timeseries
        retry = 0

        while "continuationToken" in list(response.keys()):
            headers = {
                "x-ms-client-application-name": self.ts_applicationName,
                "Authorization": f"Bearer {authorizationToken}",
                "Content-Type": "application/json",
                "cache-control": "no-cache",
                "x-ms-continuation": response["continuationToken"],
            }
            while retry < 50:
                try:
                    jsonResponse = requests.request(
                        "POST",
                        url,
                        data=json.dumps(payload),
                        headers=headers,
                        params=querystring,
                    )
                    jsonResponse.raise_for_status()
                    break
                except requests.exceptions.HTTPError:
                    error = json.loads(jsonResponse.text)
                    if(error["error"]["code"] == "TooManyRequests"):
                        # Too many requests at the same time. Waiting for a bit before retrying."
                        time.sleep(random.randint(1, 5))
                        retry += 1
                        continue
                    else:
                        raise Exception(
                            "The request to the TSI api returned an unsuccessfull status code.")

            response = json.loads(jsonResponse.text)
            result["timestamps"].extend(response["timestamps"])
            for i, property in enumerate(response["properties"]):
                result["properties"][i]["values"].extend(property["values"])
        retry = 0
        return result

    def build_dataframe_from_processes(self, processes, dTypes):
        if processes is None or len(processes) == 0:
            return pd.DataFrame()

        lambdaData = []
        dfColumns = ["timestamp", "spatialID", "objectID"]
        for task in as_completed(processes):
            result = task.result()
            if "timestamps" in result:
                for i, (timestamp) in enumerate(result["timestamps"]):
                    lambdaData.append({ "timestamp": timestamp,
                                        "spatialID": result["timeseries"][0],
                                        "objectID": result["timeseries"][1],
                                        "properties": []})
                    properties = []
                    for (j, property) in enumerate(result["properties"]):
                        if property["name"] == "spatialId" or property["name"] == "objectId":
                            continue
                        properties.append({
                            "name": property["name"],
                            "value": property["values"][i],
                        })
                        if not property["name"] in dfColumns:
                            dfColumns.append(property["name"])
                    lambdaData[len(lambdaData)-1]["properties"] = properties

        dataframe = pd.DataFrame(columns=dfColumns)
        arDf = []
        for i, (event) in enumerate(lambdaData):
            values = [pd.to_datetime(event["timestamp"]), event["spatialID"], event["objectID"]]
            indexes = ['timestamp', 'spatialID', 'objectID']
            for _, (property) in enumerate(event["properties"]):
                if(property["name"] in indexes):
                    # Received multiple different values for the exact same date and property. Update the value to the latest data we received.
                    values[indexes.index(property["name"])] = property["value"]
                else :
                    values.append(property["value"])
                    indexes.append(property["name"])
            arDf.append(pd.DataFrame([values], columns=indexes))
        dataframe = pd.concat([dataframe, pd.concat(arDf)])

        # Force dtype to what has been requestion in the properties object when calling a search, if any.
        for dtype in enumerate(dTypes):
            name = dtype[1]["name"]
            type = dtype[1]["type"]
            if name in dataframe.columns and type.lower() == 'double':
                dataframe[name] = dataframe[name].astype('float64')
        return dataframe
# endregion GetTimeSeriesData



import voxelfarm
import pandas as pd
import numpy as np
from datetime import datetime

class property_buffer_type:
    Buffer = 'buffer'
    Dataframe = 'dataframe'
    Voxel = 'voxels'

class special_properties:
    Merge = 'voxelfarm_merge'
    BufferType = 'voxelfarm_buffer_type'
    Compress = 'voxelfarm_compress'
    FixSelfIntersections = 'voxelfarm_fix_self_intersections'

class operation_type:
    sum = 0
    min = 1
    max = 2
    avg = 3

class type:
    view = 1 << 1
    project = 1 << 2
    voxel_terrain = 1 << 3
    voxel_operations = 1 << 4
    block_model = 1 << 5
    point_cloud = 1 << 6
    voxel_generator = 1 << 7
    point_cloud_raw = 1 << 8
    heightmap_raw = 1 << 9
    block_model_raw = 1 << 10
    mesh_raw = 1 << 11
    mesh = 1 << 12
    ortho = 1 << 13
    program = 1 << 14
    folder = 1 << 15
    raw_density = 1 << 16
    indexed_density = 1 << 17
    material_tracking = 1 << 18
    voxel_mesh = 1 << 19
    drill_holes = 1 << 20
    drill_holes_raw = 1 << 21
    indexed_ortho = 1 << 22
    voxel_points = 1 << 23
    voxel_plane = 1 << 24
    wms_ortho = 1 << 25
    voxel_ortho = 1 << 26
    export = 1 << 27
    report = 1 << 28
    raw_geo_chem = 1 << 29
    geo_chem = 1 << 30
    wmts_ortho = 1 << 31
    terrain_composite = 1 << 32
    value = 1
    set = 2

class attribute:
    none = 0
    volume = 1
    material = 1 << 1
    color = 1 << 2
    vector = 1 << 3
    normal = 1 << 4
    tangent = 1 << 5
    uv = 1 << 6
    color_difference = 1 << 7
    voxel_index = 1 << 8
    faces = 1 << 9

class attribute_merge_mode:
    none = 0
    add = 1
    multiply = 2
    min = 3
    max = 4
    average = 5
    unit = 6
    grade = 7

class layer_merge_mode:
    none = 0
    min = 1
    max = 2
    override = 3

class DrillHoleOptions:
    show_collars = True
    show_cylinders = True

class report_lambda_framework:

    def init(self):
        pass

    def get_voxel_size(self):
        return 0.5

    def input(self, id, label, default = 0.0):
        return 0.0

    def input_entity(self, id, label, type):
        return ''

    def input_attributes(self, id, label, type, entity):
        return ''

    def input_attribute(self, id, label, type, entity):
        return ''

    def input_query(self, id, label, entity):
        return ''

    def input_region(self, id, label):
        return ''

    def input_date(self, id, label, default = datetime.now):
        return datetime.fromtimestamp(0) 

    def input_drillhole_options(self, entity, default):
        result = DrillHoleOptions()
        return result    

    def input_string(self, id, label, default = ""):
        return ''

    def input_bool(self, id, label, default = None):
        return 0

    def set_query(self, entity, query):
        pass

    def set_attributes(self, entity, attributes):
        pass

    def set_attribute_gradient(self, entity, attribute, min_value, max_value):
        pass

    def start_composite(self):
        pass

    def add_layer(self, layers, entity, merge_mode=0):
        pass

    def get_entity_type(self, entity):
        return 0

    def get_entity_name(self, entity):
        return ''

    def get_entity_date(self, entity):
        return datetime.fromtimestamp(0) #we lose the milliseconds

    def folder_contains(self, folder, entity):
        return 0

    def get_attributes(self, attributes):
        count = 0
        return range(0, count)

    def get_attribute_name(self, attributes, index):
        return ''

    def get_attribute_index(self, attributes, name):
        return 0

    def start_sum(self, item, property):
        return -1

    def sum(self, id, value):
        pass

    def start_minimum(self, item, property):
        return -1

    def min(self, id, value):
        pass

    def start_maximum(self, item, property):
        return -1

    def max(self, id, value):
        pass

    def start_average(self, item, property):
        return -1

    def avg(self, id, value):
        pass

    def start_operation(self, item, property, operation):
        pass

    def operation(self, id, value):
        pass

    def get_material(self, entity, voxel):
        return 0

    def get_volume(self, entity, voxel):
        return 0.0

    def get_channel_volume(self, component, voxel, channel):
        return 0.0

    def get_attribute_value(self, entity, attribute, voxel):
        return 0.0

    def get_channel_attribute_value(self, component, attribute, voxel, channel):
        return 0.0

    def load_voxels(self, entity, attributes, custom_attributes):
        return 0

    def voxels_intersection(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def voxels_union(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def voxels_complement(self, componentA, componentB, attributes, custom_attributes):
        return 0

    def new_region_scope(self, region):
        return 0

    def set_field(self, v, value):
        pass

    def log(self, message):
        print(message)

    def input_duplicate(self, entity):
        return 0

    def new_plane(self, a, b, c, d):
        return -1    

    def get_enum_attribute_name_count(self, component, attribute_name):
        return 0

    def get_enum_attribute_name_value(self, component, attribute_name, index):
        return 0
        
    def get_enum_attribute_count(self, component, attribute):
        return 0

    def get_enum_attribute_value(self, component, attribute, index):
        return 0

    def set_component_attribute_merge_mode(self, component, attribute, merge_mode):
        pass
        
    def get_component_attributes(self, component):
        return []

    def get_component_data_buffers(self, component):
        return None

    def get_entity_count(self):
        return 0

    def is_meta(self):
        return True

    def add_property(self, layer, key, value):
        pass

    def store(self, component):
        pass

    def get_voxel_centroid(self, v, axis):
        return 0.0

class report_lambda_host:

    version = (3,0,1,7)

    def __init__(self, framework = None):  
        if framework:
            self.lambda_framework = framework
        else:
            if voxelfarm.voxelfarm_framework:
                self.lambda_framework = voxelfarm.voxelfarm_framework
            else:
                self.lambda_framework = report_lambda_framework()

        self.entityCount = self.lambda_framework.get_entity_count()
        self.entities = range(0, self.entityCount)

        if self.lambda_framework.is_meta():
            self.voxelCount = 1   
            self.fieldCount = 1  
        else:
            self.voxelCount = 40*40*40
            self.fieldCount = 45*45*45

        self.voxels = range(0, self.voxelCount)
        self.field = range(0, self.fieldCount)

    def get_voxels(self, component):
        voxel_count = self.lambda_framework.get_voxels(component)
        return range(0, voxel_count)

    def get_voxel_size(self):
        return self.lambda_framework.get_voxel_size()

    def input(self, id, label, default = 0.0):
        return self.lambda_framework.input(id, label, default)

    def input_entity(self, id, label, type):
        return self.lambda_framework.input_entity(id, label, type)

    def input_attributes(self, id, label, type, entity):
        return self.lambda_framework.input_attributes(id, label, type, entity)

    def input_attribute(self, id, label, type, entity):
        return self.lambda_framework.input_attributes(id, label, type, entity)

    def input_query(self, id, label, entity):
        return self.lambda_framework.input_query(id, label, entity)

    def input_region(self, id, label):
        return self.lambda_framework.input_region(id, label)

    def input_drillhole_options(self, entity, default):
        result = DrillHoleOptions()
        show_collars = 1
        show_cylinders = 1
        if default != None:
            show_collars = 1 if default.show_collars else 0
            show_cylinders = 1 if default.show_cylinders else 0
        self.lambda_framework.input_drillhole_options(entity, show_collars, show_cylinders)
        return result

    def input_date(self, id, label, default = datetime.now):
        timestamp = self.lambda_framework.input_date(id, label, 0)
        return datetime.fromtimestamp(timestamp/1000) # we lose the milliseconds

    def input_string(self, id, label, default = ""):
        return self.lambda_framework.input_string(id, label, default)

    def input_bool(self, id, label, default = None):
        return self.lambda_framework.input_bool(id, label, default)

    def set_query(self, entity, query):
        self.lambda_framework.set_query(entity, query)

    def set_attributes(self, entity, attributes):
        self.lambda_framework.set_attributes(entity, attributes)

    def set_attribute_gradient(self, entity, attribute, min_value, max_value):
        self.lambda_framework.set_attribute_gradient(entity, attribute, min_value, max_value)

    def start_composite(self):
        return self.lambda_framework.start_composite()

    def add_layer(self, layers, entity, merge_mode=layer_merge_mode.override):
        self.lambda_framework.add_layer(layers, entity, merge_mode)

    def get_entity_type(self, entity):
        return self.lambda_framework.get_entity_type(entity)

    def get_entity_name(self, entity):
        return self.lambda_framework.get_entity_name(entity)

    def get_entity_date(self, entity):
        timestamp = self.lambda_framework.get_entity_date(entity)
        return datetime.fromtimestamp(timestamp/1000) #we lose the milliseconds

    def get_entity_attributes(self, entity):
        attributes = self.lambda_framework.get_entity_attributes(entity)
        return attributes

    def folder_contains(self, folder, entity):
        return self.lambda_framework.folder_contains(folder, entity)

    def get_attributes(self, attributes):
        count = self.lambda_framework.get_attribute_count(attributes)
        return range(0, count)

    def get_attribute_name(self, attributes, index):
        return self.lambda_framework.get_attribute_name(attributes, index)

    def get_attribute_index(self, attributes, name):
        return self.lambda_framework.get_attribute_index(attributes, name)

    def start_sum(self, item, property):
        return self.start_operation(item, property, operation_type.sum)

    def sum(self, id, value):
        self.operation(id, value)

    def start_minimum(self, item, property):
        return self.start_operation(item, property, operation_type.min)

    def min(self, id, value):
        self.operation(id, value)

    def start_maximum(self, item, property):
        return self.start_operation(item, property, operation_type.max)

    def max(self, id, value):
        self.operation(id, value)

    def start_average(self, item, property):
        return self.start_operation(item, property, operation_type.avg)

    def avg(self, id, value):
        self.operation(id, value)

    def start_operation(self, item, property, operation):
        return self.lambda_framework.start_operation(item, property, operation)

    def operation(self, id, value):
        self.lambda_framework.operation(id, value)

    def get_material(self, component, voxel):
        return self.lambda_framework.get_material(component, voxel)

    def get_channels(self, component):
        count = self.lambda_framework.get_channel_count(component)
        return range(0, count)

    def get_volume(self, component, voxel):
        return self.lambda_framework.get_volume(component, voxel)

    def get_channel_volume(self, component, voxel, channel):
        return self.lambda_framework.get_volume_channel(component, voxel, channel)

    def get_attribute_value(self, component, attribute, voxel):
        return self.lambda_framework.get_attribute(component, attribute, voxel)

    def get_channel_attribute_value(self, component, attribute, voxel, channel):
        return self.lambda_framework.get_attribute_channel(component, attribute, voxel, channel)

    def load_voxels(self, entity, attributes, custom_attributes):
        return self.lambda_framework.load_voxels(entity, attributes, voxelfarm.serialize_attributes(custom_attributes))

    def voxels_intersection(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.voxels_intersection(componentA, componentB, attributes, voxelfarm.serialize_attributes(custom_attributes))

    def voxels_union(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.voxels_union(componentA, componentB, attributes, voxelfarm.serialize_attributes(custom_attributes))

    def voxels_complement(self, componentA, componentB, attributes, custom_attributes):
        return self.lambda_framework.voxels_complement(componentA, componentB, attributes, voxelfarm.serialize_attributes(custom_attributes))

    def new_region_scope(self, region):
        return range(0, self.lambda_framework.new_region_scope(region))

    def get_voxel_origin(self, v):
        return (0,0,0)

    def get_voxel_centroid(self, component, v):
        x = self.lambda_framework.get_voxel_centroid(component, v, 0)    
        y = self.lambda_framework.get_voxel_centroid(component, v, 1)    
        z = self.lambda_framework.get_voxel_centroid(component, v, 2)    
        return (x,y,z)

    def get_field_origin(self, v):
        return (0,0,0)

    def set_field(self, v, value):
        self.lambda_framework.set_field(v, value)

    def log(self, message):
        self.lambda_framework.log(message)

    def input_duplicate(self, entity):
        return self.lambda_framework.input_duplicate(entity)

    def new_plane(self, a, b, c, d):
        return self.lambda_framework.new_plane(a, b, c, d)

    def get_enum_attribute_name_count(self, component, attribute_name):
        return self.lambda_framework.get_enum_attribute_count(component, attribute_name)

    def get_enum_attribute_name_value(self, component, attribute_name, index):
        return self.lambda_framework.get_enum_attribute_name_value(component, attribute_name, index)
        
    def get_enum_attribute_count(self, component, attribute):
        return self.lambda_framework.get_enum_attribute_count(component, attribute)

    def get_enum_attribute_value(self, component, attribute, index):
        return self.lambda_framework.get_enum_attribute_value(component, attribute, index)
        
    def set_component_attribute_merge_mode(self, component, attribute, merge_mode):
        return self.lambda_framework.set_component_attribute_merge_mode(component, attribute, merge_mode)

    def valid_volume_channel(self, component, channel):
        return self.lambda_framework.valid_volume_buffer(component, channel)

    def valid_attribute_channel(self, component, attribute, channel):
        return self.lambda_framework.valid_attribute_buffer(component, attribute, channel)

    def get_voxel_volume(self):
        return self.lambda_framework.get_voxel_volume()
    
    def convert_series_to_alphanumerical(self, metadata_component, attribute, values):
        return values.apply(lambda x: self.get_enum_attribute_name_value(metadata_component, attribute, int(x)))
    
    def convert_to_alphanumerical(self, metadata_component, column, df):
        if not df.empty:
            if column:
                if column in df:
                    metadata_column = column
                    if metadata_column == 'mesh_id':
                        metadata_column = 'mesh_name'

                    return df.apply(lambda row : self.get_enum_attribute_name_value(metadata_component, metadata_column, int(row[column])-1), axis = 1)
            else:            
                metadata_column = 'mesh_name'
                return df.apply(lambda row : self.get_enum_attribute_name_value(metadata_component, metadata_column, int(row.name)-1), axis = 1)

        return np.nan

    def load_pandas_dataframe(self, component):
        component_data = []
        component_columns = self.lambda_framework.get_component_attributes(component)
        attrCount = len(component_columns)
        component_buffers = self.lambda_framework.get_component_data_buffers(component)
        if component_buffers:
            serialData = np.array(component_buffers)
            component_data = serialData.reshape(-1, attrCount)
        
        df = pd.DataFrame(data=component_data, columns=component_columns)
        if 'volume' in df:
            volume = self.get_voxel_volume()
            df['volume'] *= volume

        return df

    def sum_pandas_result(self, name, df):
        if not df.empty: 
            if isinstance(df, pd.DataFrame):
                indices = []
                if isinstance(df.index, pd.core.indexes.multi.MultiIndex):
                    for index_list in df.index:
                        multi_index = ''
                        for index in index_list:
                            if multi_index:
                                multi_index += ';' + str(index)
                            else:
                                multi_index = str(index)

                        indices.append(multi_index)
                else:
                    indices = df.index.astype(str).tolist()

                columns = df.columns.tolist()
                data = df.to_numpy().tolist()
                self.lambda_framework.sum_data_buffers(name, indices, columns, data)
            elif isinstance(df, pd.Series):
                indices = df.index.astype(str).tolist()
                columns = None
                data = df.to_numpy().tolist()
                self.lambda_framework.sum_data_buffers(name, indices, columns, data)

    def add_property(self, item, key, value):
        self.lambda_framework.add_property(item, key, value)

    def store(self, component):
        self.lambda_framework.store(component)



import requests
import os
import os.path
import glob
import datetime
import time
from random import *
import csv 
import pandas as pd
import json
import base64
import voxelfarm
from voxelfarm import voxelfarmclient

def _handle_non_serializable(obj):
    return None

class state:
    initialize = 0
    receive_data = 1
    stage_done = 2

class request:
    def __init__(self, framework, vf):
        framework.log('Creating request...')

        self.vf_api = vf
        self.lambda_framework = framework

        self.project_id = framework.input_string('project_id', 'Project Id', '')
        self.version_folder_id = framework.input_string('version_folder_id', 'Version Folder', '')

        self.update_type = framework.update_type()
        self.stage_done = framework.stage_done()

        self.state = state.initialize
        if self.version_folder_id:
            if self.stage_done:
                self.state = state.stage_done
            else:
                self.state = state.receive_data

        self.crs = {}
        # Get the coordinate system of the project
        framework.log('Retrieving project CRS...')
        result = self.vf_api.get_project_crs(self.project_id)
        if result.success:
            self.crs = result.crs
            framework.log('Retrieved project CRS')

        framework.log(f'state {self.state}')
        framework.log(f'project_id {self.project_id}')

        self.properties = {}
        self.active_version_folder_id = "0"
        self.active_version_properties = {}

        if self.state == state.initialize:
            self.raw_entity_id = None
            self.product_id = None
            self.product_folder_id = None
            self.user = None
            self.comment = None
            self.capture_date = None
            self.entity_id = None
        else:
            self.scrap_folder = framework.get_scrap_folder()          

            entity = framework.get_entity()
            self.entity_id = entity['ID']

            framework.log(f'entity_id {self.entity_id}')

            self.raw_entity_id = framework.input_string('raw_entity_id', 'Entity Id', '')
            self.product_id = framework.input_string('product_id', 'Product Id', '')
            self.product_folder_id = framework.input_string('product_folder_id', 'Product Folder', '')
            self.user = framework.input_string('user', 'User', '')
            self.comment = framework.input_string('comment', 'Comment', '')
            self.capture_date = framework.input_string('capture_date', 'Capture Date', '')

            framework.log(f'scrap_folder {self.scrap_folder}')
            framework.log(f'raw_entity_id {self.raw_entity_id}')
            framework.log(f'product_id {self.product_id}')
            framework.log(f'product_folder_id {self.product_folder_id}')
            framework.log(f'version_folder_id {self.version_folder_id}')
            framework.log(f'user {self.user}')
            framework.log(f'comment {self.comment}')
            framework.log(f'capture_date {self.capture_date}')
            framework.log(f'update_type {self.update_type}')
            framework.log(f'stage_done {self.stage_done}')

            product_entity = self.vf_api.get_entity(self.product_folder_id, self.project_id)
            framework.log(f'product_entity:{product_entity}')

            # Load previously active version properties
            framework.log('Loading properties from active version...')

            if (product_entity and 'version_active' in product_entity and product_entity['version_active'] != '0'):
                self.active_version_folder_id = product_entity['version_active']
                active_version_entity = self.vf_api.get_entity(self.active_version_folder_id, self.project_id)
                if active_version_entity:
                    for prop in active_version_entity:
                        if prop.find('version_') == 0:
                            value = active_version_entity[prop]
                            prop_name = prop.replace('version_', '', 1)
                            self.active_version_properties[prop_name] = value

            # Load version properties
            framework.log('Loading properties from current version...')
            version_entity = self.vf_api.get_entity(self.version_folder_id, self.project_id)
            if version_entity:
                for prop in version_entity:
                    if prop.find('version_') == 0:
                        value = version_entity[prop]
                        prop_name = prop.replace('version_', '', 1)
                        prop_value = version_entity[prop]
                        self.properties[prop_name] = prop_value
                        framework.log(f'{prop_name} = {prop_value}')

    def get_client_api(self):
        return self.vf_api        

    def get_entity_property(self, entity_id, property):
        self.lambda_framework.log(f'getting {entity_id} property:{property}...')

        entity = self.vf_api.get_entity(entity_id, self.project_id)
        if entity:
            self.lambda_framework.log('entity loaded')
            extended_prop = 'version_' + property
            if extended_prop in entity:
                return entity[extended_prop]

        return None

    def get_product_property(self, product_id, property):
        self.lambda_framework.log(f'getting {product_id} property:{property}...')

        workflow_entity = self.vf_api.get_entity(self.project_id, self.project_id)

        field_id = f'workflow_folder_{product_id}'
        if workflow_entity and (field_id in workflow_entity):
            product_folder_id = workflow_entity[field_id]
            self.lambda_framework.log(f'product_folder_id:{product_folder_id}')

            product_entity = self.vf_api.get_entity(product_folder_id, self.project_id)
            if product_entity:
                self.lambda_framework.log(f'product_entity:{product_entity}')
                active_version_id = product_entity['version_active']
                active_version = self.vf_api.get_entity(active_version_id, self.project_id)
                self.lambda_framework.log(f'active_version:{active_version}')
                extended_prop = 'version_' + property
                self.lambda_framework.log(f'extended_prop:{extended_prop}')
                if active_version != None and extended_prop in active_version:
                    return active_version[extended_prop]

        return None

    def get_product_singleton(self, product_id, singleton_id):
        self.lambda_framework.log(f'getting {singleton_id} for {product_id}...')

        workflow_entity = self.vf_api.get_entity(self.project_id, self.project_id)

        field_id = f'workflow_folder_{product_id}'
        if workflow_entity and (field_id in workflow_entity):
            product_folder_id = workflow_entity[field_id]
            self.lambda_framework.log(f'product_folder_id:{product_folder_id}')
            
            product_entity = self.vf_api.get_entity(product_folder_id, self.project_id)
            if product_entity:
                self.lambda_framework.log(f'product_entity: {product_entity}')
                singleton_property = f'workflow_singleton_{singleton_id}'
                if product_entity and singleton_property in product_entity:
                    return product_entity[singleton_property]

        return None

    def define_product_alias(self, alias, value):
        self.lambda_framework.log(f'Define alias {alias} for product {self.product_id}')
        updatedProperties = {}
        alias_property_id = 'alias_' + alias
        updatedProperties[alias_property_id] = value

        self.vf_api.update_entity(
            project=self.project_id,
            id=self.version_folder_id,
            fields=updatedProperties
        )

    def get_callback(self, update_type : str):
        if update_type:
            return f'{self.entity_id}/{update_type}'
        
        return None

class workflow_lambda_framework:

    def __init__(self):  
        pass

    def input_string(self, id, label, default = ""):
        return ""

    def log(self, message):
        pass

    def progress(self, progress, message):
        pass

    def get_entity(self, id = None):
        return None

    def download_entity_file(self, filename, id = None):
        return ""
    
    def get_scrap_folder(self):
        return ""

    def get_tools_folder(self):
        return ""

    def get_entity_folder(self, id = None):
        return ""

    def download_entity_files(self, id = None):
        return ""

    def download_entity_file(self, filename, id = None):
        return ""

    def attach_file(self, filename, id = None):
        return True

    def attach_folder(self, folder, id = None):
        return True

    def remove_file(self, filename, id = None):
        return True

    def upload(self, filename, name, id = None):
        pass

    def set_exit_code(self, code):
        pass

    def get_entity_file_list(self, id = None):
        return []
    
    def update_type(self):
        return None

    def stage_done(self):
        return False

    def get_vf_api(self):
        return None

    def increment_counter(self, counter, offset = 1):
        return 0
    
    def decrement_counter(self, counter, offset = 1):
        return 0

    def set_counter(self, counter, value = 0):
        pass

    def get_working_dir(self):
        return ''

    def get_notify_emails(self):
        return ''

    def get_notify_users_url(self):
        return ''

    def get_workflow_api(self):
        return ''

    def publish_feed(self, filename):
        return {'success': True}

class workflow_lambda_host:

    def __init__(self, framework = None):  
        if framework:
            self.lambda_framework = framework
        else:
            if voxelfarm.voxelfarm_framework:
                self.lambda_framework = voxelfarm.voxelfarm_framework
            else:
                self.lambda_framework = workflow_lambda_framework()

        self.vf_api = self.lambda_framework.get_vf_api()
        if self.vf_api:
            if type(self.vf_api) == str:
                self.lambda_framework.log(f'vf_api: {self.vf_api}')
                self.vf_api = voxelfarmclient.rest(self.vf_api)
        else:    
            self.lambda_framework.log('vf_api: localhost')
            self.vf_api = voxelfarmclient.rest('http://localhost')

    def input_string(self, id, label, default = ""):
        return self.lambda_framework.input_string(id, label, default)

    def log(self, message):
        self.lambda_framework.log(message)

    def progress(self, progress, message):
        self.lambda_framework.progress(progress, message)

    def get_scrap_folder(self):
        return self.lambda_framework.get_scrap_folder()

    def get_tools_folder(self):
        return self.lambda_framework.get_tools_folder()

    def get_entity_folder(self, id = None):
        return self.lambda_framework.get_entity_folder(id)

    def download_entity_files(self, id = None):
        return self.lambda_framework.download_entity_files(id)

    def download_entity_file(self, filename, id = None):
        return self.lambda_framework.download_entity_file(filename, id)

    def attach_file(self, filename, id = None):
        self.lambda_framework.attach_file(filename, id)

    def attach_folder(self, folder, id = None):
        self.lambda_framework.attach_folder(folder, id)

    def remove_file(self, filename, id = None):
        self.lambda_framework.remove_file(filename, id)

    def upload(self, filename, name, id = None):
        self.lambda_framework.Upload(filename, name, id)

    def set_exit_code(self, code):
        self.lambda_framework.set_exit_code(code)

    def get_entity(self, id = None):
        return self.lambda_framework.get_entity(id)

    def get_entity_file_list(self, id = None):
        return self.lambda_framework.get_entity_file_list(id)

    def update_type(self):
        if self.lambda_framework.stage_done():
            return self.lambda_framework.update_type()
        
        return None

    def stage_done(self):
        return self.lambda_framework.stage_done()

    def increment_counter(self, counter, offset = 1):
        return self.lambda_framework.increment_counter(counter, offset)
    
    def decrement_counter(self, counter, offset = 1):
        return self.lambda_framework.decrement_counter(counter, offset)

    def set_counter(self, counter, value = 0):
        self.lambda_framework.set_counter(counter, value)

    def get_working_dir(self):
        return self.lambda_framework.get_working_dir()
    
    def get_notify_emails(self):
        return self.lambda_framework.get_notify_emails()

    def get_notify_users_url(self):
        return self.lambda_framework.get_notify_users_url()

    def get_workflow_api(self):
        return self.lambda_framework.get_workflow_api()

    def find_product_definition(self, product, workflow):
        if 'id' in workflow:
            if workflow['id'] == product:
                return workflow
            else:
                if 'tracks' in workflow:
                    for track in workflow['tracks']:
                        self.lambda_framework.log(f'track: {track}')

                        product_definition = self.find_product_definition(product, track)
                        if product_definition:
                            return product_definition

        return None

    def _is_number(self, element) -> bool:
        try:
            float(element)
            return True
        except:
            return False

    def _create_workflow_folders(self, vf, workflow, project_id, project_entity, parent_folder, fields, crs):
        result = vf.api_result()

        if 'singletons' in workflow:
            existing_version = vf.get_entity(parent_folder, project_id)
            existing_singleton_ids = {}
            if existing_version:
                for prop in existing_version:
                    if prop.startswith('workflow_singleton_'):
                        existing_singleton_ids[prop] = existing_version[prop]

            singleton_ids = {}
            for singleton in workflow['singletons']:
                create_singleton = False    
                singleton_key = f'workflow_singleton_{singleton}'
                if singleton_key in existing_singleton_ids:
                    singleton_id = existing_singleton_ids[singleton_key]
                    singleton_entity = vf.get_entity(singleton_id, project_id)
                    if singleton_entity == None:
                        create_singleton = True    
                else:
                    create_singleton = True    
                    
                if create_singleton:    
                    self.lambda_framework.log(f'Creating Product singleton {workflow["id"]} {parent_folder} {singleton}...')
                    api_result = vf.create_entity_raw(
                        project=project_id, 
                        type='RAW', 
                        name=f'{singleton}', 
                        fields={
                            'file_folder': parent_folder,
                            'capture_date' : str(1000 * int(time.time()))
                        },
                        crs=crs)
                    
                    if not api_result.success:
                        return api_result
                    
                    singleton_id = api_result.id
                    singleton_ids[f'workflow_singleton_{singleton}'] = singleton_id
                    self.lambda_framework.log(f'Created Raw Singleton {singleton_id}')

            if singleton_ids:
                vf.update_entity(
                    project=project_id, 
                    id=parent_folder, 
                    fields=singleton_ids)

        if 'tracks' in workflow:
            for track in workflow['tracks']:
                product_id = track['id']
                field_id = f'workflow_folder_{product_id}'
                if field_id in project_entity:
                    folder_id = project_entity[field_id]
                    folder = vf.get_entity(folder_id, project_id)
                    if folder != None:
                        api_result = self._create_workflow_folders(vf, track, project_id, project_entity, folder_id, fields, crs)
                        if not api_result.success:
                            return api_result
                        
                        continue
                    else:        
                        self.lambda_framework.log(f'folder does not exists: {folder}')    
                else:
                    self.lambda_framework.log(f'field_id does not exists: {field_id}')    

                track_name = track['name']
                self.lambda_framework.log(f'Creating Product folder {product_id} {parent_folder} {track_name}...')
                api_result = vf.create_folder(
                    project=project_id, 
                    folder=parent_folder,
                    name=track['name'],
                    fields={
                        'workflow_product' : product_id,
                        'version_last' : '0',
                        'version_active' : '0'
                    })
                
                if not api_result.success:
                    return api_result
                
                sub_folder_id = api_result.id

                fields[field_id] = sub_folder_id
                self.lambda_framework.log('Created Product folder')
                api_result = self._create_workflow_folders(vf, track, project_id, project_entity, sub_folder_id, fields, crs)
                if not api_result.success:
                    return api_result
        
        result.success = True
        return result    

    def set_workflow_definition(self, workflow_definition):
        self.lambda_framework.log('Create workflow request')

        result = {
            'success': False,
            'error_info': '',
            'complete': False}

        workflow_request = request(self.lambda_framework, self.vf_api)

        if workflow_request.state == state.initialize:
            self.lambda_framework.log('Initialize workflow')
            project_entity = self.vf_api.get_entity(workflow_request.project_id, workflow_request.project_id)

            encoded_workflow = json.dumps(workflow_definition, default=_handle_non_serializable)
            encoded_bytes = encoded_workflow.encode('ascii')
            encoded_bytes = base64.b64encode(encoded_bytes)
            encoded_workflow = encoded_bytes.decode('ascii')

            fields = {
                'workflow' : encoded_workflow
            }

            self.lambda_framework.log('Creating Workflow Folders')
            api_result = self._create_workflow_folders(self.vf_api, workflow_definition, workflow_request.project_id, project_entity, '0', fields, workflow_request.crs)

            if api_result.success:
                self.lambda_framework.log(fields)
                self.lambda_framework.log(f'Saving changes to project {workflow_request.project_id}...')

                api_result = self.vf_api.update_entity(
                    project=workflow_request.project_id, 
                    id=workflow_request.project_id, 
                    fields=fields)

                if api_result.success:
                    result['success'] = True
                    result['complete'] = True
                else:        
                    self.lambda_framework.log('Error updating project.')
                    result['error_info'] = api_result.error_info
            else:        
                self.lambda_framework.log('Creating workflow folder failed.')
                result['error_info'] = api_result.error_info
        else:
            self.lambda_framework.log(f'Find product definition: {workflow_request.product_id}')
            product_definition = self.find_product_definition(workflow_request.product_id, workflow_definition)
            if product_definition:
                if workflow_request.state == state.stage_done:
                    if 'on_stage_done' in product_definition:
                        self.lambda_framework.log('Execute on_stage_done event')
                        event_result = product_definition['on_stage_done'](self.vf_api, workflow_request, self)

                        if event_result:
                            if 'success' in event_result:
                                if event_result['success']:
                                    result = event_result
                                    self.lambda_framework.log('Done for on_stage_done event')
                                else:    
                                    if 'error_info' in event_result:
                                        self.lambda_framework.log('Error in on_stage_done event: ' + result['error_info'])
                                    else:
                                        result['error_info'] = 'Unknown error'
                            else:
                                result['error_info'] = 'on_stage_done success result not found'
                        else:        
                            result['error_info'] = 'on_stage_done return no data'
                    else:
                        result['error_info'] = 'on_stage_done event not defined'
                elif workflow_request.state == state.receive_data:  
                    if 'on_receive_data' in product_definition:
                        self.lambda_framework.log('Execute on_receive_data event')
                        event_result = product_definition['on_receive_data'](self.vf_api, workflow_request, self)

                        if event_result:
                            if 'success' in event_result:
                                if event_result['success']:
                                    self.lambda_framework.log('Done for on_receive_data event')
                                    result = event_result
                                else:    
                                    if 'error_info' in event_result:
                                        result['error_info'] = event_result['error_info']
                                    else:
                                        result['error_info'] = 'Unknown error'
                            else:
                                result['error_info'] = 'on_receive_data success result not found'
                        else:
                            result['error_info'] = 'on_receive_data return no data'
                    else:
                        result['error_info'] = 'on_receive_data event not defined'
                else:
                    self.lambda_framework.log('Unknown workflow state')
                    result['error_info'] = 'Unknown workflow state'
                    return result

                # write properties
                self.lambda_framework.log(f'Saving properties...{workflow_request.project_id}...{workflow_request.version_folder_id}')
                updatedProperties = {}
                for prop in workflow_request.properties:
                    version_property_id = 'version_' + prop
                    property_value = workflow_request.properties[prop]
                    updatedProperties[version_property_id] = property_value
                    self.lambda_framework.log(f'{version_property_id} = {property_value}')

                api_result = self.vf_api.update_entity(
                    project=workflow_request.project_id,
                    id=workflow_request.version_folder_id,
                    fields=updatedProperties  
                )   

                if api_result.success:
                    self.lambda_framework.log('Done saving properties')
                else:
                    self.lambda_framework.log(f'Error saving properties: {api_result.error_info}')
            else:
                self.lambda_framework.log('Product definition not found')
                result['error_info'] = 'Product definition not found'

            self.lambda_framework.log('Done set_workflow_definition')

        if 'success' in result and result['success']:
            if "complete" in result and result["complete"]:
                # Auto-activate version
                if (workflow_request.state != state.initialize) and ('activated' not in workflow_request.properties):
                    self.lambda_framework.log(f'Updating product version_active folder...{workflow_request.product_folder_id}')
                    product_entity = self.lambda_framework.get_entity(workflow_request.product_folder_id)

                    if product_entity and product_entity.ContainsKey('version_active'):
                        previous_active_version = product_entity['version_active'] 
                        if previous_active_version == '':
                            previous_active_version = '0'
                        self.lambda_framework.log(f'Previously active version: {previous_active_version}')
                        api_result = self.vf_api.update_entity(
                            project=workflow_request.project_id,
                            id=workflow_request.product_folder_id,
                            fields={
                                'version_active' : str(workflow_request.version_folder_id)
                            }
                        )

                        if not api_result.success:
                            result['error_info'] = api_result.error_info

                        workflow_request.properties['activated'] = '1'
                        self.lambda_framework.log('Updated product active_version.')        
                    else:
                        self.lambda_framework.log(f'product_entity: {workflow_request.product_folder_id} not found')

                self.lambda_framework.log("workflow_lambda Done")
                self.lambda_framework.set_exit_code(0)
            else:
                self.lambda_framework.log("workflow_lambda Partial")
        else:
            self.lambda_framework.log("workflow_lambda Error: " + result['error_info'])
            self.lambda_framework.set_exit_code(1)

    def get_product_singleton(self, product_id, singleton_id):
        self.lambda_framework.log(f'getting {singleton_id} for {product_id}...')

        entity = self.lambda_framework.get_entity()
        if entity and entity.ContainsKey('project'):
            project_id = entity['project']
            project_entity = self.lambda_framework.get_entity(project_id)

            field_id = f'workflow_folder_{product_id}'
            if project_entity and project_entity.ContainsKey(field_id):
                product_folder_id = project_entity[field_id]
                self.lambda_framework.log(f'product_folder_id:{product_folder_id}')

                if product_entity:
                    product_entity = self.vf_api.get_entity(product_folder_id)
                    self.lambda_framework.log(f'product_entity: {product_entity}')

                    singleton_property = f'workflow_singleton_{singleton_id}'
                    if product_entity and product_entity.ContainsKey(singleton_property):
                        return product_entity[singleton_property]
        else:
            self.lambda_framework.log('Project not found in entity.')

        return None

    def get_product_property(self, product_id, property):
        self.lambda_framework.log(f'getting {product_id} property:{property}...')

        entity = self.lambda_framework.get_entity()
        if entity and entity.ContainsKey('project'):
            project_id = entity['project']
            project_entity = self.lambda_framework.get_entity(project_id)

            if project_entity and project_entity.ContainsKey('workflow_folder_' + product_id):
                product_folder_id = project_entity['workflow_folder_' + product_id]
                product_entity = self.lambda_framework.get_entity(product_folder_id)
                if product_entity and product_entity.ContainsKey('version_active'):
                    active_version_id = product_entity['version_active']
                    active_version = self.lambda_framework.get_entity(active_version_id)
                    extended_prop = 'version_' + property
                    if active_version and active_version.ContainsKey(extended_prop):
                        return active_version[extended_prop]
                    else:
                        self.lambda_framework.log('Problem with entity active_version ' + active_version_id)
                else:
                    self.lambda_framework.log('Problem with entity product_entity ' + product_folder_id)
            else:
                self.lambda_framework.log('Problem with entity project ' + project_id)
        else:
            self.lambda_framework.log('Project not found in entity.')

        return None

    def get_parameter_dataframe(self, product_id):
        attribute_product = self.get_product_property(product_id, 'report_entity')
        if attribute_product:
            attribute_file = self.lambda_framework.download_entity_file('report.csv', attribute_product)
            if os.path.isfile(attribute_file):
                return pd.read_csv(attribute_file)
            else:
                self.lambda_framework.log('Parameter file not found.')
        else:
            self.lambda_framework.log('Attribute product not found.')

        return pd.DataFrame()

    def load_feed(self, workflow_request: request, entity_id:str):
        self.lambda_framework.log(f'load_feed {entity_id}...')
        entity_path = self.download_entity_files(entity_id)
        if not os.path.isdir(entity_path):
            self.lambda_framework.log('Error Raw product empty')
            return {'success': False}

        self.lambda_framework.log(f'Raw Entity loaded successfully {entity_path}')
        
        types = ('*.ftr', 'report.csv')
        entity_files = []
        for type in types:
            files = glob.glob(entity_path + "\\" + type)
            if files:
                entity_files.extend(files)

        self.lambda_framework.log(f'entity_files:{entity_files}')

        for file_name in entity_files:
            entity_file = os.path.join(workflow_request.scrap_folder, file_name)
            self.lambda_framework.log(f'feed file:{entity_file}')
            if entity_file:
                if file_name.endswith('.ftr'):
                    df_model = pd.read_feather(entity_file)
                elif file_name.endswith('.csv'):
                    df_model = pd.read_csv(entity_file)
                else:
                    self.lambda_framework.log(f'Unsupported file type {file_name}')
                    return {'success': False}
                if not df_model.empty:
                    return {'success': True, 'df': df_model}
                else:           
                    self.lambda_framework.log(f'file {file_name} is empty.')
                    return {'success': False}
                
        return {'success': True, 'df': None}

    def publish_reports(self, workflow_request: request, report_name:str, report_id: str, constant_spatial_id: str, constant_object_id: str, propertyIds:tuple, timestampID: str = None, spatialId: str = None, objectId: str = None):
        self.lambda_framework.log(f'Publish reports calling timeseries')
        self.lambda_framework.log(f'report_id:{report_id}')

        propertyStr = ''
        for val in propertyIds:
            propertyStr += ','.join(map(str, val)) + "|"

        self.lambda_framework.log(f'propertyStr:{propertyStr}')
        self.lambda_framework.log(f'Calling process entity to publish report {report_id}...')
        mesh_entity = self.lambda_framework.get_entity(report_id)
        if not mesh_entity:
            self.lambda_framework.log(f'Error getting mesh_entity {report_id}')
            return {'success': False, 'error_info': 'Error getting mesh_entity'}
        file_date = mesh_entity['file_date']
        self.lambda_framework.log(f'mesh_entity.file_date:{file_date}')
        iso_time = datetime.datetime.fromtimestamp(int(file_date)/1000.0).isoformat()
        self.lambda_framework.log(f'mesh_entity.iso_time:{iso_time}')
        result = self.load_feed(workflow_request, report_id)
        if not 'success' in result and result['success']:
            return {'success': False, 'error_info': 'load_feed failed'}
        
        if 'df' in result and result['df'] is not None and not result['df'].empty:
            self._handle_dataframe(workflow_request, result['df'], report_id, f"{iso_time}", timestampID, spatialId, objectId, propertyIds, constant_spatial_id, constant_object_id)
            self.lambda_framework.log(f'End process for report {report_id}')    
        else:
            self.lambda_framework.log(f'result[\'df\'] is not defined or empty for report {report_id}')
        return {'success': True}

    def _handle_dataframe(self, workflow_request: request, df_model:pd.DataFrame, report_id: str, timestamp:str, timestampId:str, spatialId: str, objectId: str, propertyIds:tuple, constant_spatial_id:str, constant_object_id:str):
        self.lambda_framework.log(f'handle_dataframe:timestamp:{timestamp}|timestampId:{timestampId}|spatialId:{spatialId}|objectId:{objectId}|constant_spatial_id:{constant_spatial_id}|constant_object_id:{constant_object_id}')
        self.lambda_framework.log(df_model.dtypes.to_string())
        self.lambda_framework.log(df_model.head().to_string())
        if timestampId in df_model or timestamp is not None:
            eventsToSend = []
            for idx, row in df_model.iterrows():
                valid = False   
                spatial_value = '-'
                if spatialId and spatialId in row:
                    spatial_value = row[spatialId]
                    valid = True
                elif constant_spatial_id and constant_object_id:
                    valid = True
                
                object_value = '-'
                if objectId and objectId in row:
                    object_value = row[objectId]
                    valid = True

                if valid:
                    if timestampId:
                        try:
                            timestamp = str(row[timestampId])
                            if timestamp == 'None' or timestamp.strip() == '':
                                self.lambda_framework.log(f'Mandatory timestamp value is empty for {row}. Please check the input file. Skipping this row.')
                                continue
                            if " " in timestamp:
                                if "AM" or "PM" in timestamp:
                                    entry_date = time.mktime(datetime.datetime.strptime(timestamp, "%d/%m/%Y %H:%M:%S %p").timetuple())
                                else: 
                                    entry_date = time.mktime(datetime.datetime.strptime(timestamp, "%Y-%m-%d %H:%M:%S").timetuple())
                            else:
                                entry_date = time.mktime(datetime.datetime.strptime(timestamp, "%Y-%m-%d").timetuple())
                            iso_time = datetime.datetime.fromtimestamp(entry_date).isoformat()
                        except:
                            self.lambda_framework.log(f'Something went wrong converting the date for {timestamp}. Row:{row}. Please check the input file. Skipping this row.')
                            continue
                    else:
                        iso_time = timestamp

                    eventObj = {
                        "time": iso_time,
                        "spatialId": constant_spatial_id if constant_spatial_id != '' else spatial_value, 
                        "objectId": constant_object_id if constant_object_id != '' else object_value,
                    }
                    validPropertyValueFound = False
                    for propertyId in propertyIds:
                        if propertyId[0] in df_model:
                            validPropertyValueFound = True
                            eventObj[propertyId[1]] = float(row[propertyId[0]]) if self._is_number(row[propertyId[0]]) else (row[propertyId[0]] if not pd.isnull(row[propertyId[0]]) else '')

                    if validPropertyValueFound:
                        eventsToSend.append(eventObj)
                        #if idx % 250 == 0:
                            #self.lambda_framework.log(f'event to send JSON :{json.dumps(eventObj)}')
                            
            self.lambda_framework.log(f'eventsToSend count:{len(eventsToSend)}')
            if len(eventsToSend) > 0:
                self.lambda_framework.log(f'first event from the list to send to Time Series: {json.dumps(eventsToSend[0])}')

            filename = f"dump_tsi_{report_id}.csv"
            df_dump = pd.DataFrame(eventsToSend) 
            df_dump.to_csv(os.path.join(workflow_request.scrap_folder, filename), index=False)
            self.lambda_framework.publish_feed(filename)
        else:
            self.lambda_framework.log(timestampId + ' column not found in dataframe or "timestamp" value was empty (' + timestamp + ').')

    def _get_file_path(self, file):
        if os.path.exists(file):
            return file
        else:
            script_dir = self.lambda_framework.get_working_dir()
            self.lambda_framework.log('file is a relative path')
            file_path = os.path.join(script_dir, file)

            if os.path.exists(file_path):
                return file_path
    
        return None

    def _load_file(self, file):
        file_path = self._get_file_path(file)

        if os.path.exists(file_path):
            return open(file_path)
    
        return None

    def create_view(self, workflow_request : request, name : str, view_type : str, view_lambda : str, inputs : dict, props:dict):
        self.lambda_framework.log(f'create_view:name:{name}|view_type:{view_type}|view_lambda:{view_lambda}|inputs:{inputs}|props:{props}')

        if view_type == None:
            lambda_file = self._load_file(view_lambda)
            if lambda_file == None:
                return {'success': False, 'error_info': 'Lambda file not found'}

            result = self.vf_api.create_lambda_python(
                project=workflow_request.project_id, 
                type=self.vf_api.lambda_type.View,
                name=name, 
                fields={
                    'file_folder': workflow_request.version_folder_id,
                    'virtual': '1'
                },
                code=lambda_file.read())
            if not result.success:
                return {'success': False, 'error_info': result.error_info}
            view_type = result.id

        input_fields = {
                'file_folder' : workflow_request.version_folder_id,
                'view_type' : view_type,
                'virtual' : '1',
                'state' : 'COMPLETE',
                'color_legend_attribute' : '',
                'color_legend_attribute_index' : '-1',
                'color_legend_gradient' : 'isoluminant_cgo_70_c39_n256',
                'color_legend_interpolate_gradient' : '1',
                'color_legend_mode' : '2',
                'color_legend_range_max' : '100',
                'color_legend_range_min' : '0',
                'color_legend_range_step' : '1',
                'color_legend_reverse_gradient' : '0',
                'file_date' : str(1000 * int(time.time())),
                'file_type' : 'VIEW',
                'input_filter_colors' : '0',
                'input_filter_e' : '8',
                'input_filter_normals' : '0',
                'input_label_colors' : 'Use Ortho-imagery',
                'input_label_e' : 'Terrain',
                'input_label_normals' : 'Use high resolution detail',
                'input_type_colorlegend' : '7',
                'input_type_colors' : '6',
                'input_type_e' : '3',
                'input_type_normals' : '6',
                'input_value_colors' : '0',
                'input_value_normals' : '0',         
            }
            
        for key in inputs:
            input_fields['input_value_' + key] = inputs[key]

        for key in props:
            input_fields[key] = props[key]
            
        result = self.vf_api.create_entity_raw(
            project=workflow_request.project_id,
            type=self.vf_api.entity_type.View,
            name=name,
            fields=input_fields,
            crs={}
        )

        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        view_object = result.id
        
        self.lambda_framework.log(f'created_view:name:{name}|view_object:{view_object}')

        result = self.vf_api.create_entity_raw(
            project=workflow_request.project_id,
            type=self.vf_api.entity_type.View,
            name=name,
            fields={
                'file_folder' : workflow_request.version_folder_id,
                'view_type' : 'container',
                'state' : 'COMPLETE',
                'entity_container' : view_object
            },
            crs={}
        )

        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        return {'success': True, 'id': result.id, 'error_info': 'None'}

    def create_report(self, workflow_request : request, name : str, report_lambda : str, region : str, lod : int, inputs : dict, fields : dict = None, update_type : str = None):
        report_lambda_id = report_lambda
        if report_lambda.endswith('.py'):
            lambda_file = self._load_file(report_lambda)
            if lambda_file == None:
                return {'success': False, 'error_info': 'Lambda file not found'}
            
            result = self.vf_api.create_lambda_python(
                project=workflow_request.project_id, 
                type=self.vf_api.lambda_type.Report,
                name=f"Lambda for: {name}", 
                fields={
                    'virtual': '1',
                    'file_folder': workflow_request.version_folder_id
                },
                code=lambda_file.read())
            
            if not result.success:
                return {'success': False, 'error_info': result.error_info}
            
            report_lambda_id = result.id

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id

        if update_type:
            fields['callback_update_type'] = workflow_request.get_callback(update_type)

        result = self.vf_api.create_report(
            project=workflow_request.project_id, 
            program=report_lambda_id, 
            region=region,
            lod=str(lod),
            name=name, 
            fields=fields,
            inputs=inputs)
        
        if result.success:
            if update_type:
                self.increment_counter(f'update_type_{update_type}')
        else:    
            return {'success': False, 'lambda_id': report_lambda_id, 'error_info': result.error_info}
        
        return {'success': True, 'id': result.id, 'lambda_id': report_lambda_id, 'error_info': 'None'}

    def create_lambda(self, workflow_request : request, name : str, type : str, lambda_code : str):
        lambda_file = self._load_file(lambda_code)
        if lambda_file == None:
            return {'success': False, 'error_info': 'Lambda file not found'}

        result = self.vf_api.create_lambda_python(
            project=workflow_request.project_id, 
            type=type,
            name=name, 
            fields={
                'file_folder': workflow_request.version_folder_id
            },
            code=lambda_file.read())
        
        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        
        report_lambda_id = result.id
        return {'success': True, 'id': report_lambda_id}

    def create_export(self, workflow_request : request, name : str, export_lambda : str, region : str, lod : int, inputs : dict, fields : dict = None, update_type : str = None):
        lambda_file = self._load_file(export_lambda)
        if lambda_file == None:
            return {'success': False, 'error_info': 'Lambda file not found'}

        result = self.vf_api.create_lambda_python(
            project=workflow_request.project_id, 
            type=self.vf_api.lambda_type.Report,
            name="Export Lambda for " + name, 
            fields={
                'virtual': '1',
                'file_folder': workflow_request.version_folder_id
            },
            code=lambda_file.read())
        if not result.success:
            return {'success': False, 'error_info': result.error_info}
        report_lambda_id = result.id

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id
        fields['export_type'] = 'mesh'

        if update_type:
            fields['callback_update_type'] = workflow_request.get_callback(update_type)

        result = self.vf_api.create_export(
            project=workflow_request.project_id, 
            program=report_lambda_id, 
            region=region,
            lod=str(lod),
            name=name, 
            fields=fields,
            inputs=inputs)

        if result.success:
            if update_type:
                self.increment_counter(f'update_type_{update_type}')
        else:
            return {'success': False, 'lambda_id': report_lambda_id, 'error_info': result.error_info}

        return {'success': True, 'id': result.id, 'lambda_id': report_lambda_id, 'error_info': 'None'}

    def get_meta_attributes(self, workflow_request : request, projectId:str):
        attribute_product = workflow_request.get_product_property('META_ATTR_BM', 'report_entity')
        self.lambda_framework.log(f'attribute_product:{attribute_product}')
        file_name = 'report.csv'
        local_attr_file = os.path.join(workflow_request.scrap_folder, file_name)
        self.lambda_framework.log(f'Downloading {file_name} to {local_attr_file}...')
        if not self.vf_api.download_file(projectId, attribute_product, file_name, local_attr_file):
            return {'success': False, 'error_info': f'Could not download {file_name} to {local_attr_file}'}
        self.lambda_framework.log(f'Downloaded {file_name} to {local_attr_file}...')
        
        self.lambda_framework.log(f'Processing {local_attr_file}...')
        
        csvfile = open(f"{local_attr_file}", mode='r', encoding='utf-8-sig' )
        reader = csv.DictReader(csvfile)
        record_count = 0
        view_attribute = ''
        attributes_arr = []
        groups_arr = []
        grades_arr = []
        for row in reader:
            #self.lambda_framework.log(f'row:{row}')
            ID = row['Property']
            Report = row['Report']
            View = row['View']
            Group = row['Group']
            Grade = row['Grade']
            if View == '1' and view_attribute == '':
                self.lambda_framework.log(f'Assigning attribute: {ID} with index: {record_count}')
                view_attribute = ID
            if Report == '1':
                attributes_arr.append(ID)
            if Group == '1':
                groups_arr.append(ID)
            if Grade == '1':
                grades_arr.append(ID)
            record_count += 1

        attributes = ",".join(attributes_arr)
        groups = ",".join(groups_arr)
        grades = ",".join(grades_arr)

        self.lambda_framework.log(f'attributes:{attributes}')
        self.lambda_framework.log(f'groups:{groups}')
        self.lambda_framework.log(f'grades:{grades}') 

        return {'attributes':attributes, 'groups': groups, 'grades':grades, 'view_attribute': view_attribute}
        
    def notify_users(self, file_entity_id):
        self.lambda_framework.log(f'notify_users for file_entity_id: {file_entity_id}')
        file_entity = self.lambda_framework.get_entity(file_entity_id)
        if file_entity != None:
            emails_to = self.lambda_framework.get_notify_emails()
            if emails_to is None:
                self.lambda_framework.log('No email were sent because OS getenv variable vf_notify_emails is None')
                return { "success": True }
            
            self.lambda_framework.log(f'emails_to: {emails_to}')

            emails_to = emails_to.split(",")
            folder_entity = self.lambda_framework.get_entity(file_entity['file_folder'])
            parent_entity = self.lambda_framework.get_entity(folder_entity['file_folder'])
            
            if 'user' in file_entity and file_entity['user'] not in emails_to:
                emails_to.append(file_entity['user'])
            if 'user' in folder_entity and folder_entity['user'] not in emails_to:
                emails_to.append(folder_entity['user'])
            if 'user' in parent_entity and parent_entity['user'] not in emails_to:
                emails_to.append(parent_entity['user'])

            self.lambda_framework.log(f'notify users: {emails_to}')

            url = self.lambda_framework.get_notify_users_url()
            self.lambda_framework.log(f'notify url: {url}')

            if file_entity["state"] == "COMPLETE":
                subject = f"Completed {file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})"
                body = f"Dear team member,\n\nThe following item has been completed:\n\n    {file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})\n\nYou can access this item here:\n\n{self.lambda_framework.get_workflow_api()}app/catalog/{file_entity_id}\n\nPlease do not reply to this message. It has been generated by an automated workflow."
            else:
                subject = f"Error in {file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})"
                body = f"Dear team member,\n\nWe have encountered some issues while processing the following item:    \n\n{file_entity['name']} in {parent_entity['name']} ({folder_entity['name']})\n\n For more information, please access the log tab of the item. You can access this item here:\n\n{self.lambda_framework.get_workflow_api()}app/catalog/{file_entity_id}\n\nPlease do not reply to this message. It has been generated by an automated workflow."

            self.lambda_framework.log(f'subject: {subject}')
            self.lambda_framework.log(f'body: {body}')

            data = {
                "from": "noreply@voxelfarm.com",
                "to": ",".join(emails_to),
                "subject": subject,
                "body": body
            }
            headers = {"Content-type": "application/json", "Accept": "text/plain"}

            req = requests.post(url, data=json.dumps(data), headers=headers)
            if req.status_code >= 200 and req.status_code < 300:
                self.lambda_framework.log(f'Emails sent successfully')
                return { "success": True }
            else:
                self.lambda_framework.log(f'Error sending the emails')
                return { "success": False }
        else:
            self.lambda_framework.log(f'Entity : {file_entity_id} not found')
            return { "success": False }

    def process_entity(self, workflow_request, type, name, fields = None, update_type = None):

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id

        if update_type:
            fields['callback_update_type'] = workflow_request.get_callback(update_type)

        api_result = self.vf_api.create_entity_processed(
            project=workflow_request.project_id, 
            type=type, 
            name=name, 
            fields=fields, 
            crs=workflow_request.crs, 
            callback=None)
        
        if update_type and api_result.success:
            self.increment_counter(f'update_type_{update_type}')
        
        return api_result

    def process_lambda_entity(self, workflow_request, name, inputs, code, files, fields = None, update_type = None):
        verified_files = []
        if files:
            for file in files:
                verified_files.append(self._get_file_path(file))

        if fields == None:
            fields = {}

        fields['file_folder'] = workflow_request.version_folder_id

        if update_type:
            fields['callback_update_type'] = workflow_request.get_callback(update_type)

        api_result =  self.vf_api.create_process_entity(
            project=workflow_request.project_id, 
            name=name, 
            fields=fields, 
            inputs=inputs,
            code=code,
            files=verified_files, 
            callback=None)
    
        if update_type and api_result.success:
            self.increment_counter(f'update_type_{update_type}')

        return api_result    

    def create_product_version(self, project_id, product_id, inputs = None, files = None):
        self.lambda_framework.log(f'Retrieving project entity  {project_id}')
        project_entity = self.vf_api.get_entity(project_id, project_id)
        if project_entity:
            # Get the coordinate system of the project
            self.lambda_framework.log('Retrieving project CRS...')
            api_result = self.vf_api.get_project_crs(project_id)
            if not api_result.success:
                self.lambda_framework.log(api_result.error_info)
                return None

            crs = api_result.crs

            if 'workflow_id' not in project_entity:
                self.lambda_framework.log('Error_info: Workflow id not found')
                return None
            
            workflow_id = project_entity['workflow_id']
            self.lambda_framework.log(f'workflow_id: {workflow_id}')

            field_id = f'workflow_folder_{product_id}'
            if field_id not in project_entity:
                self.lambda_framework.log('Error_info: Workflow product not found')
                return None

            product_folder_id = project_entity[field_id]

            product_entity = self.vf_api.get_entity(product_folder_id, project_id)
            if not product_entity:
                self.lambda_framework.log('Error_info: Product entity not found')
                return None

            self.lambda_framework.log(f'product_entity:{product_entity}')

            # Get new version number for product
            last_version = int(product_entity['version_last']) if 'version_last' in product_entity else 0
            version_number = last_version + 1
            self.lambda_framework.log(f'Updating version {version_number} in product')
            self.vf_api.update_entity(
                project=project_id,
                id=product_folder_id,
                fields={
                    'version_last' : str(version_number)
                }
            )
            self.lambda_framework.log(f'Updated version in product')

            now = datetime.datetime.now()
            version_label = now.strftime("%d/%m/%Y %H:%M:%S:%f") 

            capture_date = ''
            user = ''
            comment = ''
            if inputs and 'capture_date' in inputs:
                capture_date = inputs['capture_date']

            if inputs and 'user' in inputs:
                user = inputs['user']

            if inputs and 'comment' in inputs:
                comment = inputs['comment']
            
            self.lambda_framework.log('Creating Version folder...')
            api_result = self.vf_api.create_folder(
                project=project_id, 
                folder=product_folder_id,
                name=f'Version {version_number}',
                fields={
                    'workflow_product' : product_id,
                    'workflow_id' : workflow_id,
                    'version_label' : version_label,
                    'version_number' : f'{version_number}',
                    'capture_date' : capture_date,
                    'user': user,
                    'comment': comment
                })
            
            if not api_result.success:
                self.lambda_framework.log(api_result.error_info)
                return None

            version_folder_id = api_result.id

            # Create raw object
            self.lambda_framework.log('Create raw entity')
            api_result = self.vf_api.create_entity_raw(
                project=project_id, 
                type='RAW', 
                name=f'New {product_id}', 
                fields={
                    'file_folder': version_folder_id,
                    'capture_date' : str(1000 * int(time.time())),
                    'user': user,
                    'comment': comment
                },
                crs=crs)
            
            if not api_result.success:
                self.lambda_framework.log(api_result.error_info)
                return None
            
            entity_id = api_result.id
            self.lambda_framework.log(f'Created Raw Object {entity_id}')

            for file in files:
                self.lambda_framework.attach_file(file, entity_id)

            if not inputs:
                inputs = {}

            inputs['raw_entity_id'] = entity_id
            inputs['product_folder_id'] = product_folder_id
            inputs['version_folder_id'] = version_folder_id
            inputs['project_id'] = project_id
            inputs['product_id'] = product_id

            apiResult = self.vf_api.create_workflow_entity(
                project=project_id,
                name="Workflow Lambda",
                fields={
                    'file_folder' : version_folder_id
                },
                inputs=inputs)

            if apiResult.success:
                return version_number
        else:
            self.lambda_framework.log('Project not found')

        return None





